# -*- coding: UTF-8 -*-
# File: concurrency.py
# Author: Yuxin Wu <ppwwyyxx@gmail.com>
# Credit belongs to Xinyu Zhou
from __future__ import unicode_literals

import threading
import multiprocessing
import atexit
import bisect
from contextlib import contextmanager
import signal
import weakref
import six
from datetime import datetime
import time
import random
import os, sys
import subprocess
from six.moves import queue
import time
from . import logger

__all__ = ['StoppableThread', 'LoopThread','TimerThread', 'ensure_proc_terminate',
           'OrderedResultGatherProc', 'OrderedContainer', 'DIE',
           'mask_sigint', 'start_proc_mask_signal', 'StoppableProcess', 'GatherMultiProcs']


class StoppableThread(threading.Thread):
  """
  A thread that has a 'stop' event.
  """
  def __init__(self, evt=None):
    super(StoppableThread, self).__init__()
    if evt is None:
      evt = threading.Event()
    self._stop_evt = evt

  def stop(self):
    """ stop the thread"""
    self._stop_evt.set()

  def stopped(self):
    """ check whether the thread is stopped or not"""
    return self._stop_evt.isSet()

  def queue_put_stoppable(self, q, obj):
    """ put obj to queue, but will give up if the thread is stopped"""
    while not self.stopped():
      try:
        q.put(obj, timeout=5)
        break
      except queue.Full:
        pass

  def queue_get_stoppable(self, q):
    """ take obj from queue, but will give up if the thread is stopped"""
    while not self.stopped():
      try:
        return q.get(timeout=5)
      except queue.Empty:
        pass
  
  
class LoopThread(StoppableThread):
  """ A pausable thread that simply runs a loop"""
  def __init__(self, func, pausable=True):
    """
    :param func: the function to run
    """
    super(LoopThread, self).__init__()
    self._func = func
    self._pausable = pausable
    if pausable:
        self._lock = threading.Lock()
    self.daemon = True

  def run(self):
    while not self.stopped():
        if self._pausable:
            self._lock.acquire()
            self._lock.release()
        self._func()

  def pause(self):
    assert self._pausable
    self._lock.acquire()

  def resume(self):
    assert self._pausable
    self._lock.release()


class TimerThread(StoppableThread):
  ''' A timer thread that run func periodically'''
  def __init__(self,func,periodic = 10):
    '''
    :param func:
    :param periodical: unit is sec
    '''
    super(TimerThread, self).__init__()
    self._func = func
    self.daemon = True
    self.periodic = periodic

  def run(self):
    while not self.stopped():
      if type(self._func) == list:
        for func in self._func:
            func()
      else:
        self._func()
      time.sleep(self.periodic)


class DIE(object):
  """ A placeholder class indicating end of queue """
  pass


def ensure_proc_terminate(proc):
  if isinstance(proc, list):
    for p in proc:
        ensure_proc_terminate(p)
    return

  def stop_proc_by_weak_ref(ref):
    proc = ref()
    if proc is None:
      return
    if not proc.is_alive():
      return
    proc.terminate()
    proc.join()

  assert isinstance(proc, multiprocessing.Process)
  atexit.register(stop_proc_by_weak_ref, weakref.ref(proc))


@contextmanager
def mask_sigint():
  sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
  yield
  signal.signal(signal.SIGINT, sigint_handler)


def start_proc_mask_signal(proc):
  if not isinstance(proc, list):
    proc = [proc]

  with mask_sigint():
    for p in proc:
      p.start()


def subproc_call(cmd, timeout=None):
  try:
    output = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT,
            shell=True, timeout=timeout)
    return output
  except subprocess.TimeoutExpired as e:
    logger.warn("Command timeout!")
    logger.warn(e.output)
  except subprocess.CalledProcessError as e:
    logger.warn("Commnad failed: {}".format(e.returncode))
    logger.warn(e.output)


class OrderedContainer(object):
  """
  Like a priority queue, but will always wait for item with index (x+1) before producing (x+2).
  """
  def __init__(self, start=0):
    self.ranks = []
    self.data = []
    self.wait_for = start

  def put(self, rank, val):
    idx = bisect.bisect(self.ranks, rank)
    self.ranks.insert(idx, rank)
    self.data.insert(idx, val)

  def has_next(self):
    if len(self.ranks) == 0:
      return False
    return self.ranks[0] == self.wait_for

  def get(self):
    assert self.has_next()
    ret = self.data[0]
    rank = self.ranks[0]
    del self.ranks[0]
    del self.data[0]
    self.wait_for += 1
    return rank, ret


class OrderedResultGatherProc(multiprocessing.Process):
  """
  Gather indexed data from a data queue, and produce results with the
  original index-based order.
  """
  def __init__(self, data_queue, nr_producer, start=0):
    """
    :param data_queue: a multiprocessing.Queue to produce input dp
    :param nr_producer: number of producer processes. Will terminate after receiving this many of DIE sentinel.
    :param start: the first task index
    """
    super(OrderedResultGatherProc, self).__init__()
    self.data_queue = data_queue
    self.ordered_container = OrderedContainer(start=start)
    self.result_queue = multiprocessing.Queue()
    self.nr_producer = nr_producer

  def run(self):
    nr_end = 0
    try:
      while True:
        task_id, data = self.data_queue.get()
        if task_id == DIE:
          self.result_queue.put((task_id, data))
          nr_end += 1
          if nr_end == self.nr_producer:
            return
        else:
          self.ordered_container.put(task_id, data)
          while self.ordered_container.has_next():
            self.result_queue.put(self.ordered_container.get())
    except Exception as e:
      import traceback
      traceback.print_exc()
      raise e

  def get(self):
    return self.result_queue.get()


class StoppableProcess(multiprocessing.Process):
  def __init__(self, queue_size=-1):
    super(StoppableProcess, self).__init__()
    self._is_stop = False
    self._condition = None
    self._queue = multiprocessing.Queue(queue_size)

  def stop(self):
    """ stop the thread"""
    self._is_stop = True

  def stopped(self):
    """ check whether the thread is stopped or not"""
    return self._is_stop
  
  @property
  def process_condition(self):
    if self._condition == None:
      self._condition = multiprocessing.Condition()
    return self._condition
  
  @property
  def process_queue(self):
    return self._queue
  
  
class GatherMultiProcs(object):
  @staticmethod
  def process_func(data_flow, data_pipe, condition):
    # 1.step prepare random seed for current pid
    seed = (os.getpid() +
            int(datetime.now().strftime("%Y%m%d%H%M%S%f"))) % 4294967295
    random.seed(seed)
    
    while True:
      # 2.step put in queue
      for data in data_flow.iterator_value():
        data_pipe.put(data)
      
      # 3.step add DIE flag
      with condition:
        data_pipe.put(DIE)
        condition.wait()

  def __init__(self, data_flow, nr=2, cache=2):
    self._is_running = False
    self._nr = nr
    self._queue = multiprocessing.Queue(nr * cache)
    self._condition = multiprocessing.Condition()
    self._processes = [multiprocessing.Process(target=GatherMultiProcs.process_func,
                       args=(data_flow, self._queue, self._condition)) for _ in range(nr)]

    for p in self._processes:
      p.daemon = True
  
  def iterator_value(self):
    assert(self._queue.empty())
    
    # launch all processes
    if not self._is_running:
      for p in self._processes:
        p.start()
      self._is_running = True
    
    # notify all waiting processes
    with self._condition:
      self._condition.notify_all()

    while True:
      data = self._queue.get()
      if data == DIE:
        DIE_nr = 1
        # waiting until nr DIE
        while DIE_nr < self._nr:
          temp = self._queue.get()
          if temp == DIE:
            DIE_nr += 1
          else:
            yield temp
          
        raise StopIteration
      yield data


# -*- coding: UTF-8 -*-
# @Time    : 17-6-8
# @File    : distorted.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.core import *
from antgo.codebook.bboxes.bboxes import *
import random
import math


class Distorted(Node):
  def __init__(self, inputs,aspect_ratio_range=(0.6, 1.67), area_range=(0.1, 1.0), max_try_times=100):
    super(Distorted, self).__init__(name=None, action=self.action, inputs=inputs)

    self._aspect_ratio_range = aspect_ratio_range
    self._area_range = area_range
    self._max_try_times = max_try_times
  
  def _sample_distorted_bounding_box_like(self, image_size,
                                         bounding_boxes,
                                         aspect_ratio_range=(0.6, 1.67),
                                         area_range=(0.1, 1.0),
                                         max_try_times=100):
    width = image_size[1]
    height = image_size[0]

    box_num = bounding_boxes.shape[0]
    random_box_index = int(np.floor(random.random() * box_num))
    focus_box = bounding_boxes[random_box_index, :]
    focus_box_x0 = focus_box[0] / float(width)
    focus_box_y0 = focus_box[1] / float(height)
    focus_box_x1 = focus_box[2] / float(width)
    focus_box_y1 = focus_box[3] / float(height)
    focus_box_w = focus_box_x1 - focus_box_x0
    focus_box_h = focus_box_y1 - focus_box_y0
    try_time = 0
    w = 1.0
    h = 1.0
    while try_time < max_try_times:
      area = random.random() * (area_range[1] - area_range[0]) + area_range[0]
      aspect_ratio = random.random() * (aspect_ratio_range[1] - aspect_ratio_range[0]) + aspect_ratio_range[0]

      h = math.sqrt((float(width) * area) / (float(height) * aspect_ratio))
      w = area / h

      if w >= focus_box_w and h >= focus_box_h:
          break

      try_time += 1

    if try_time == max_try_times:
      w = 1.0
      h = 1.0

    random_x0_max = focus_box_x0
    random_x0_min = np.maximum(focus_box_x0 - (w - focus_box_w), 0)

    random_y0_max = focus_box_y0
    random_y0_min = np.maximum(focus_box_y0 - (h - focus_box_h), 0)

    x0 = random.random() * (random_x0_max - random_x0_min) + random_x0_min
    y0 = random.random() * (random_y0_max - random_y0_min) + random_y0_min

    begin = np.maximum(np.array((int(x0 * width), int(y0 * height), 0)), np.array((0, 0, 0)))

    w = np.minimum(1.0 - x0, w)
    h = np.minimum(1.0 - y0, h)
    size = np.minimum(np.array((int(w * width), int(h * height), -1)), np.array([width - 1, height - 1, -1]))
    return begin, size, focus_box
  
  def action(self, *args, **kwargs):
    assert(len(args) == 1)
    data, annotation = args[0] if len(args[0]) == 2 else (args[0], {})
    assert('bbox' in annotation)

    # distorted bbox
    begin, size, distoted_bbox = self._sample_distorted_bounding_box_like(data.shape,
                                                                          annotation['bbox'],
                                                                          aspect_ratio_range=(0.95, 1.0),
                                                                          area_range=(0.1, 0.3))

    # crop image
    crop_image = data[begin[1]:begin[1]+size[1], begin[0]:begin[0]+size[0]]
    # remained bboxes
    remained_bboxes,remained_bboxes_ind = \
      bboxes_filter_overlap(np.array((begin[0], begin[1], begin[0]+size[0], begin[1]+size[1])),
                            annotation['bbox'], 0.5)

    bboxes = bboxes_translate(np.array((begin[0], begin[1], begin[0]+size[0], begin[1]+size[1])),
                              remained_bboxes)


    # modify annotation
    annotation['bbox'] = bboxes
    if 'category_id' in annotation:
      annotation['category_id'] = annotation['category_id'][remained_bboxes_ind]
    if 'category' in annotation:
      annotation['category'] = [annotation['category'][i] for i in remained_bboxes_ind]
    if 'area' in annotation:
      annotation['area'] = annotation['area'][remained_bboxes_ind]
    if 'info' in annotation:
      annotation['info'] = (crop_image.shape[0], crop_image.shape[1], crop_image.shape[2])

    if 'segmentation' in annotation:
      annotation['segmentation'] = \
          [annotation['segmentation'][i][begin[1]:begin[1]+size[1], begin[0]:begin[0]+size[0]]
           for i in remained_bboxes_ind]
    annotation['cell'] = np.array((begin[0], begin[1], begin[0]+size[0], begin[1]+size[1]))

    return (crop_image, annotation)


# -*- coding: UTF-8 -*-
# @Time    : 2018/11/9 6:00 PM
# @File    : debug.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.utils import logger
from antgo.context import *
from antgo.dataflow.dataset.random_dataset import *
from antgo.utils.utils import *
from antgo.dataflow.recorder import *
import os
import sys
import yaml
from datetime import datetime
from types import FunctionType
from antgo import config
from antgo.dataflow.recorder import *
from antgo.dataflow.common import *
import traceback

def debug_training_process(dataset_func, param_config=None, dump_dir=None):
  # 0.step get global context
  ctx = get_global_context()
  ctx.debug = True

  # 1.step parse params config file
  if param_config is not None:
    logger.info('load param file %s'% param_config)
    params = yaml.load(open(param_config, 'r'))
    ctx.params = params

  # 2.step call traing process
  if dump_dir is None:
    train_time_stamp = datetime.fromtimestamp(timestamp()).strftime('%Y%m%d.%H%M%S.%f')
    logger.info('build dump folder %s'%train_time_stamp)
    dump_dir = os.path.join(os.curdir, 'dump', train_time_stamp)

  if not os.path.exists(dump_dir):
    os.makedirs(dump_dir)
  logger.info('start debug training process')

  dataset_obj = None
  if isinstance(dataset_func, FunctionType):
    dataset_obj = RandomDataset('train', '')
    dataset_obj.data_func = dataset_func
  else:
    Config = config.AntConfig
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    Config.parse_xml(config_xml)
    dataset_obj = dataset_func('train', os.path.join(Config.data_factory, dataset_func.__name__))

  ctx.call_training_process(dataset_obj, dump_dir=dump_dir)


def debug_infer_process(dataset_func, param_config=None, dump_dir=None):
  # 0.step get global context
  ctx = get_global_context()
  ctx.debug = True

  # 1.step parse params config file
  if param_config is not None:
    logger.info('load param file %s' % param_config)
    params = yaml.load(open(param_config, 'r'))
    ctx.params = params

  # 2.step call traing process
  if dump_dir is None:
    train_time_stamp = datetime.fromtimestamp(timestamp()).strftime('%Y%m%d.%H%M%S.%f')
    logger.info('build dump folder %s'%train_time_stamp)
    dump_dir = os.path.join(os.curdir, 'dump', train_time_stamp)
  if not os.path.exists(dump_dir):
    os.makedirs(dump_dir)
  logger.info('start debug infer process')
  ctx.recorder = EmptyRecorderNode()
  ctx.ant = True

  dataset_obj = None
  if isinstance(dataset_func, FunctionType):
    dataset_obj = RandomDataset('test', '')
    dataset_obj.data_func = dataset_func
    ctx.call_infer_process(dataset_obj, dump_dir=dump_dir)
  else:
    Config = config.AntConfig
    config_xml = os.path.join(os.environ['HOME'], '.config', 'antgo', 'config.xml')
    Config.parse_xml(config_xml)
    dataset_obj = dataset_func('test', os.path.join(Config.data_factory, dataset_func.__name__))
    dataset_obj = DataAnnotationBranch(Node.inputs(dataset_obj))
    ctx.recorder = RecorderNode(Node.inputs(dataset_obj.output(1)))

    ctx.call_infer_process(dataset_obj.output(0), dump_dir=dump_dir)



# -*- coding: UTF-8 -*-
# @Time    : 18-1-4
# @File    : iphone2dslr.py
# @Author  : Jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np
__all__ = ['Iphone2Dslr']


IPHONE2DSLR_URL = ''
class Iphone2Dslr(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(Iphone2Dslr, self).__init__(train_or_test, dir)
    assert(train_or_test in ['train', 'test', 'sample'])

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    # 0.step maybe download
    if not os.path.exists(os.path.join(self.dir, 'iphone2dslr_flower')):
      self.download(self.dir, file_names=[], default_url=IPHONE2DSLR_URL, auto_unzip=True)

    # 1.step extract data list
    self.data_a_list = []
    self.data_b_list = []
    if train_or_test == 'train':
      for a in os.listdir(os.path.join(self.dir, 'iphone2dslr_flower', 'trainA')):
        if a[0] != '.':
          self.data_a_list.append(os.path.join(self.dir, 'iphone2dslr_flower', 'trainA', a))
  
      for b in os.listdir(os.path.join(self.dir, 'iphone2dslr_flower', 'trainB')):
        if b[0] != '.':
          self.data_b_list.append(os.path.join(self.dir, 'iphone2dslr_flower', 'trainB', b))
    else:
      for a in os.listdir(os.path.join(self.dir, 'iphone2dslr_flower', 'testA')):
        if a[0] != '.':
          self.data_a_list.append(os.path.join(self.dir, 'iphone2dslr_flower', 'testA', a))
  
      for b in os.listdir(os.path.join(self.dir, 'iphone2dslr_flower', 'testB')):
        if b[0] != '.':
          self.data_b_list.append(os.path.join(self.dir, 'iphone2dslr_flower', 'testB', b))

    self.ids = list(range(len(self.data_a_list)))

  @property
  def size(self):
    return len(self.ids)
  
  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1
    
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)

      b_idxs = list(range(len(self.data_b_list)))
      if self.rng:
        self.rng.shuffle(b_idxs)

      for a_k, b_k in zip(idxs, b_idxs):
        a = self.data_a_list[a_k]
        b = self.data_b_list[b_k]
        a_img = imread(a)
        b_img = imread(b)
        yield [a_img, b_img]

  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    a = self.data_a_list[id]
    random_b_index = np.random.randint(0, len(self.data_b_list), 1)[0]
    b = self.data_b_list[random_b_index]
    a_img = imread(a)
    b_img = imread(b)
    return a_img, b_img

  def split(self, split_params={}, split_method=''):
    raise NotImplementedError



# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : spider_api.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.utils import logger
from antgo.ant.download import *
from antgo.ant import flags
import os
import requests
import time
import  urllib.parse
import uuid

def spider_api_baidu():
    FLAGS = flags.AntFLAGS

    datasource_name = 'baidu'
    datasource_type = ''
    datasource_keyword = ''
    for p in FLAGS.param().split(','):
        k,v = p.split(":")     
        if k == 'type':
            datasource_type = v
        elif k == 'keyword':
            datasource_keyword = v

    if datasource_name not in ['baidu', 'google', 'bing', 'vcg']:
        logger.error('Only support datasource baidu/google/bing/vcg')
        return
    
    if datasource_type not in ['image', 'video']:
        logger.error('Only support datasource type image/video')
        return
    
    if datasource_keyword == '':
        logger.error('Must set keyword')
        return

    time_stamp = (int)(time.time())
    if not os.path.exists('./spider_%d'%(time_stamp)):
        os.makedirs('./spider_%d'%(time_stamp))
    baidu_download(datasource_keyword, {'download_data_type': datasource_type}, './spider_%d'%(time_stamp))

def spider_api_bing():
    FLAGS = flags.AntFLAGS

    datasource_name = 'bing'
    datasource_type = ''
    datasource_keyword = ''
    for p in FLAGS.param().split(','):
        k,v = p.split(":")    
        if k == 'type':
            datasource_type = v
        elif k == 'keyword':
            datasource_keyword = v

    if datasource_name not in ['baidu', 'google', 'bing', 'vcg']:
        logger.error('Only support datasource baidu/google/bing/vcg')
        return
    
    if datasource_type not in ['image', 'video']:
        logger.error('Only support datasource type image/video')
        return
    
    if datasource_keyword == '':
        logger.error('Must set keyword')
        return

    time_stamp = (int)(time.time())
    if not os.path.exists('./spider_%d'%(time_stamp)):
        os.makedirs('./spider_%d'%(time_stamp))
    bing_download(datasource_keyword, {'download_data_type': datasource_type}, './spider_%d'%(time_stamp))

def spider_api_google():
    FLAGS = flags.AntFLAGS

    logger.error("In coming")

def spider_api_vcg():
    FLAGS = flags.AntFLAGS

    datasource_name = 'vcg'
    datasource_type = ''
    datasource_keyword = ''
    for p in FLAGS.param().split(','):
        k,v = p.split(":")    
        if k == 'type':
            datasource_type = v
        elif k == 'keyword':
            datasource_keyword = v

    if datasource_name not in ['baidu', 'google', 'bing', 'vcg']:
        logger.error('Only support datasource baidu/google/bing/vcg')
        return
    
    if datasource_type not in ['image', 'video']:
        logger.error('Only support datasource type image/video')
        return
    
    if datasource_keyword == '':
        logger.error('Must set keyword')
        return

    time_stamp = (int)(time.time())
    if not os.path.exists('./spider_%d'%(time_stamp)):
        os.makedirs('./spider_%d'%(time_stamp))
    vcg_download(datasource_keyword, {'download_data_type': datasource_type}, './spider_%d'%(time_stamp))



# -*- coding: UTF-8 -*-
# @Time : 26/03/2018
# @File : ali_fashion_attribute_error.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
import numpy as np


default = {'AntALIFashionAttributeError': ('FashionAttributeError', 'CLASSIFICATION')}
class AntALIFashionAttributeError(AntMeasure):
  def __init__(self, task):
    super(AntALIFashionAttributeError, self).__init__('FashionAttributeError')
    assert(task.task_type == 'CLASSIFICATION')
    self.is_support_rank = True

    self.cloth_attribs = ['skirt_length_labels',
                          'coat_length_labels',
                          'collar_design_labels',
                          'lapel_design_labels',
                          'neck_design_labels',
                          'neckline_design_labels',
                          'pant_length_labels',
                          'sleeve_length_labels']

  def eva(self, data, label):
    '''

    :param data:
    :param label:
    :return:
    '''
    # traverse all attrib
    mean_ap = []
    for cloth_attrib in self.cloth_attribs:
      max_attr_value_prob_list = []
      zip_data = data if label is None else zip(data, label)
      for predict, gt in zip_data:
        # filter
        if gt['category'] != cloth_attrib:
          continue

        predict_prob, predict_cloth_attrib = predict
        assert(predict_cloth_attrib == cloth_attrib)

        max_attr_value_prob = np.max(predict_prob)
        max_attr_value_index = np.argmax(predict_prob)

        if gt['category_id'][max_attr_value_index] == 1:
          max_attr_value_prob_list.append((max_attr_value_prob, 1))
        elif gt['category_id'][max_attr_value_index] == 0:
          max_attr_value_prob_list.append((max_attr_value_prob, 0))

      # compute AP
      sorted_prob = sorted(max_attr_value_prob_list, key=lambda x: -x[0])
      predict_correct_count = 0
      predict_count = 0
      val_list = []
      for prob in sorted_prob:
        if prob[1] == 1:
          predict_correct_count += 1
          predict_count += 1
        else:
          predict_count += 1

        val = float(predict_correct_count) / float(predict_count)
        val_list.append(val)

      if len(val_list) == 0:
        print('sdf')
      ap = np.mean(val_list)
      mean_ap.append(ap)

    mean_ap_val = np.mean(mean_ap)
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': mean_ap_val, 'type': 'SCALAR'}]}}


# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Deploy Slim models across multiple clones and replicas.

# TODO(sguada) docstring paragraph by (a) motivating the need for the file and
# (b) defining clones.

# TODO(sguada) describe the high-level components of model deployment.
# E.g. "each model deployment is composed of several parts: a DeploymentConfig,
# which captures A, B and C, an input_fn which loads data.. etc

To easily train a model on multiple GPUs or across multiple machines this
module provides a set of helper functions: `create_clones`,
`optimize_clones` and `deploy`.

Usage:

  g = tf.Graph()

  # Set up DeploymentConfig
  config = model_deploy.DeploymentConfig(num_clones=2, clone_on_cpu=True)

  # Create the global step on the device storing the variables.
  with tf.device(config.variables_device()):
    global_step = slim.create_global_step()

  # Define the inputs
  with tf.device(config.inputs_device()):
    images, labels = LoadData(...)
    inputs_queue = slim.data.prefetch_queue((images, labels))

  # Define the optimizer.
  with tf.device(config.optimizer_device()):
    optimizer = tf.train.MomentumOptimizer(FLAGS.learning_rate, FLAGS.momentum)

  # Define the model including the loss.
  def model_fn(inputs_queue):
    images, labels = inputs_queue.dequeue()
    predictions = CreateNetwork(images)
    slim.losses.log_loss(predictions, labels)

  model_dp = model_deploy.deploy(config, model_fn, [inputs_queue],
                                 optimizer=optimizer)

  # Run training.
  slim.learning.train(model_dp.train_op, my_log_dir,
                      summary_op=model_dp.summary_op)

The Clone namedtuple holds together the values associated with each call to
model_fn:
  * outputs: The return values of the calls to `model_fn()`.
  * scope: The scope used to create the clone.
  * device: The device used to create the clone.

DeployedModel namedtuple, holds together the values needed to train multiple
clones:
  * train_op: An operation that run the optimizer training op and include
    all the update ops created by `model_fn`. Present only if an optimizer
    was specified.
  * summary_op: An operation that run the summaries created by `model_fn`
    and process_gradients.
  * total_loss: A `Tensor` that contains the sum of all losses created by
    `model_fn` plus the regularization losses.
  * clones: List of `Clone` tuples returned by `create_clones()`.

DeploymentConfig parameters:
  * num_clones: Number of model clones to deploy in each replica.
  * clone_on_cpu: True if clones should be placed on CPU.
  * replica_id: Integer.  Index of the replica for which the model is
      deployed.  Usually 0 for the chief replica.
  * num_replicas: Number of replicas to use.
  * num_ps_tasks: Number of tasks for the `ps` job. 0 to not use replicas.
  * worker_job_name: A name for the worker job.
  * ps_job_name: A name for the parameter server job.

TODO(sguada):
  - describe side effect to the graph.
  - what happens to summaries and update_ops.
  - which graph collections are altered.
  - write a tutorial on how to use this.
  - analyze the possibility of calling deploy more than once.


"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import tensorflow as tf
from tensorflow.python.ops import control_flow_ops

slim = tf.contrib.slim


__all__ = ['create_clones',
           'deploy',
           'optimize_clones',
           'DeployedModel',
           'DeploymentConfig',
           'Clone',
          ]


# Namedtuple used to represent a clone during deployment.
Clone = collections.namedtuple('Clone',
                               ['outputs',  # Whatever model_fn() returned.
                                'scope',  # The scope used to create it.
                                'device',  # The device used to create.
                               ])

# Namedtuple used to represent a DeployedModel, returned by deploy().
DeployedModel = collections.namedtuple('DeployedModel',
                                       ['train_op',  # The `train_op`
                                        'summary_op',  # The `summary_op`
                                        'total_loss',  # The loss `Tensor`
                                        'clones',  # A list of `Clones` tuples.
                                       ])

# Default parameters for DeploymentConfig
_deployment_params = {'num_clones': 1,
                      'clone_on_cpu': False,
                      'replica_id': 0,
                      'num_replicas': 1,
                      'num_ps_tasks': 0,
                      'worker_job_name': 'worker',
                      'ps_job_name': 'ps'}


def create_clones(config, model_fn, args=None, kwargs=None):
  """Creates multiple clones according to config using a `model_fn`.

  The returned values of `model_fn(*args, **kwargs)` are collected along with
  the scope and device used to created it in a namedtuple
  `Clone(outputs, scope, device)`

  Note: it is assumed that any loss created by `model_fn` is collected at
  the tf.GraphKeys.LOSSES collection.

  To recover the losses, summaries or update_ops created by the clone use:
  ```python
    losses = tf.get_collection(tf.GraphKeys.LOSSES, clone.scope)
    summaries = tf.get_collection(tf.GraphKeys.SUMMARIES, clone.scope)
    update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS, clone.scope)
  ```

  The deployment options are specified by the config object and support
  deploying one or several clones on different GPUs and one or several replicas
  of such clones.

  The argument `model_fn` is called `config.num_clones` times to create the
  model clones as `model_fn(*args, **kwargs)`.

  If `config` specifies deployment on multiple replicas then the default
  tensorflow device is set appropriatly for each call to `model_fn` and for the
  slim variable creation functions: model and global variables will be created
  on the `ps` device, the clone operations will be on the `worker` device.

  Args:
    config: A DeploymentConfig object.
    model_fn: A callable. Called as `model_fn(*args, **kwargs)`
    args: Optional list of arguments to pass to `model_fn`.
    kwargs: Optional list of keyword arguments to pass to `model_fn`.

  Returns:
    A list of namedtuples `Clone`.
  """
  clones = []
  args = args or []
  kwargs = kwargs or {}
  with slim.arg_scope([slim.model_variable, slim.variable],
                      device=config.variables_device()):
    # Create clones.
    for i in range(0, config.num_clones):
      i = config.clone_id_map[i] if i in config.clone_id_map else i

      kwargs['clone'] = i
      with tf.name_scope(config.clone_scope(i)) as clone_scope:
        clone_device = config.clone_device(i)
        with tf.device(clone_device):
          with tf.variable_scope(tf.get_variable_scope(),
                                 reuse=True if i > 0 else None):
            outputs = model_fn(*args, **kwargs)
          clones.append(Clone(outputs, clone_scope, clone_device))
          
  return clones


def _gather_clone_loss(clone, num_clones, regularization_losses):
  """Gather the loss for a single clone.

  Args:
    clone: A Clone namedtuple.
    num_clones: The number of clones being deployed.
    regularization_losses: Possibly empty list of regularization_losses
      to add to the clone losses.

  Returns:
    A tensor for the total loss for the clone.  Can be None.
  """
  # The return value.
  sum_loss = None
  # Individual components of the loss that will need summaries.
  clone_loss = None
  regularization_loss = None
  # Compute and aggregate losses on the clone device.
  with tf.device(clone.device):
    all_losses = []
    clone_losses = tf.get_collection(tf.GraphKeys.LOSSES, clone.scope)
    if clone_losses:
      clone_loss = tf.add_n(clone_losses, name='clone_loss')
      if num_clones > 1:
        clone_loss = tf.div(clone_loss, 1.0 * num_clones, name='scaled_clone_loss')
      all_losses.append(clone_loss)
    if regularization_losses:
      regularization_loss = tf.add_n(regularization_losses,
                                     name='regularization_loss')
      all_losses.append(regularization_loss)
    if all_losses:
      sum_loss = tf.add_n(all_losses)
      
  # Add the summaries out of the clone device block.
  if clone_loss is not None:
    tf.summary.scalar(clone.scope + '/clone_loss', clone_loss)
  if regularization_loss is not None:
    tf.summary.scalar('regularization_loss', regularization_loss)
  return sum_loss


def _optimize_clone(optimizer, clone, num_clones, regularization_losses,
                    **kwargs):
  """Compute losses and gradients for a single clone.

  Args:
    optimizer: A tf.Optimizer  object.
    clone: A Clone namedtuple.
    num_clones: The number of clones being deployed.
    regularization_losses: Possibly empty list of regularization_losses
      to add to the clone losses.
    **kwargs: Dict of kwarg to pass to compute_gradients().

  Returns:
    A tuple (clone_loss, clone_grads_and_vars).
      - clone_loss: A tensor for the total loss for the clone.  Can be None.
      - clone_grads_and_vars: List of (gradient, variable) for the clone.
        Can be empty.
  """
  sum_loss = _gather_clone_loss(clone, num_clones, regularization_losses)
  clone_grad = None
  if sum_loss is not None:
    with tf.device(clone.device):
      clone_grad = optimizer.compute_gradients(sum_loss, **kwargs)
  return sum_loss, clone_grad


def optimize_clones(clones,
                    optimizer,
                    regularization_losses=None,
                    **kwargs):
  """Compute clone losses and gradients for the given list of `Clones`.

  Note: The regularization_losses are added to the first clone losses.

  Args:
   clones: List of `Clones` created by `create_clones()`.
   optimizer: An `Optimizer` object.
   regularization_losses: Optional list of regularization losses. If None it
     will gather them from tf.GraphKeys.REGULARIZATION_LOSSES. Pass `[]` to
     exclude them.
   **kwargs: Optional list of keyword arguments to pass to `compute_gradients`.

  Returns:
   A tuple (total_loss, grads_and_vars).
     - total_loss: A Tensor containing the average of the clone losses including
       the regularization loss.
     - grads_and_vars: A List of tuples (gradient, variable) containing the sum
       of the gradients for each variable.

  """
  grads_and_vars = []
  clones_losses = []
  num_clones = len(clones)
  if regularization_losses is None:
    regularization_losses = tf.get_collection(
        tf.GraphKeys.REGULARIZATION_LOSSES)
  
  for clone in clones:
    with tf.name_scope(clone.scope):
      clone_loss, clone_grad = _optimize_clone(
          optimizer, clone, num_clones, regularization_losses, **kwargs)
      if clone_loss is not None:
        clones_losses.append(clone_loss)
        grads_and_vars.append(clone_grad)
      # Only use regularization_losses for the first clone
      regularization_losses = None
  # Compute the total_loss summing all the clones_losses.
  total_loss = tf.add_n(clones_losses, name='total_loss')
  # Sum the gradients across clones.
  grads_and_vars = _sum_clones_gradients(grads_and_vars)
  return total_loss, grads_and_vars


def deploy(config,
           model_fn,
           args=None,
           kwargs=None,
           optimizer=None,
           summarize_gradients=False):
  """Deploys a Slim-constructed model across multiple clones.

  The deployment options are specified by the config object and support
  deploying one or several clones on different GPUs and one or several replicas
  of such clones.

  The argument `model_fn` is called `config.num_clones` times to create the
  model clones as `model_fn(*args, **kwargs)`.

  The optional argument `optimizer` is an `Optimizer` object.  If not `None`,
  the deployed model is configured for training with that optimizer.

  If `config` specifies deployment on multiple replicas then the default
  tensorflow device is set appropriatly for each call to `model_fn` and for the
  slim variable creation functions: model and global variables will be created
  on the `ps` device, the clone operations will be on the `worker` device.

  Args:
    config: A `DeploymentConfig` object.
    model_fn: A callable. Called as `model_fn(*args, **kwargs)`
    args: Optional list of arguments to pass to `model_fn`.
    kwargs: Optional list of keyword arguments to pass to `model_fn`.
    optimizer: Optional `Optimizer` object.  If passed the model is deployed
      for training with that optimizer.
    summarize_gradients: Whether or not add summaries to the gradients.

  Returns:
    A `DeployedModel` namedtuple.

  """
  # Gather initial summaries.
  summaries = set(tf.get_collection(tf.GraphKeys.SUMMARIES))

  # Create Clones.
  clones = create_clones(config, model_fn, args, kwargs)
  first_clone = clones[0]

  # Gather update_ops from the first clone. These contain, for example,
  # the updates for the batch_norm variables created by model_fn.
  update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS, first_clone.scope)

  train_op = None
  total_loss = None
  with tf.device(config.optimizer_device()):
    if optimizer:
      # Place the global step on the device storing the variables.
      with tf.device(config.variables_device()):
        global_step = slim.get_or_create_global_step()

      # Compute the gradients for the clones.
      total_loss, clones_gradients = optimize_clones(clones, optimizer)

      if clones_gradients:
        if summarize_gradients:
          # Add summaries to the gradients.
          summaries |= set(_add_gradients_summaries(clones_gradients))

        # Create gradient updates.
        grad_updates = optimizer.apply_gradients(clones_gradients,
                                                 global_step=global_step)
        update_ops.append(grad_updates)

        update_op = tf.group(*update_ops)
        with tf.control_dependencies([update_op]):
          train_op = tf.identity(total_loss, name='train_op')
    else:
      clones_losses = []
      regularization_losses = tf.get_collection(
          tf.GraphKeys.REGULARIZATION_LOSSES)
      for clone in clones:
        with tf.name_scope(clone.scope):
          clone_loss = _gather_clone_loss(clone, len(clones),
                                          regularization_losses)
          if clone_loss is not None:
            clones_losses.append(clone_loss)
          # Only use regularization_losses for the first clone
          regularization_losses = None
      if clones_losses:
        total_loss = tf.add_n(clones_losses, name='total_loss')

    # Add the summaries from the first clone. These contain the summaries
    # created by model_fn and either optimize_clones() or _gather_clone_loss().
    summaries |= set(tf.get_collection(tf.GraphKeys.SUMMARIES,
                                       first_clone.scope))

    if total_loss is not None:
      # Add total_loss to summary.
      summaries.add(tf.summary.scalar('total_loss', total_loss))

    if summaries:
      # Merge all summaries together.
      summary_op = tf.summary.merge(list(summaries), name='summary_op')
    else:
      summary_op = None

  return DeployedModel(train_op, summary_op, total_loss, clones)


def _sum_clones_gradients(clone_grads):
  """Calculate the sum gradient for each shared variable across all clones.

  This function assumes that the clone_grads has been scaled appropriately by
  1 / num_clones.

  Args:
    clone_grads: A List of List of tuples (gradient, variable), one list per
    `Clone`.

  Returns:
     List of tuples of (gradient, variable) where the gradient has been summed
     across all clones.
  """
  sum_grads = []
  for grad_and_vars in zip(*clone_grads):
    # Note that each grad_and_vars looks like the following:
    #   ((grad_var0_clone0, var0), ... (grad_varN_cloneN, varN))
    grads = []
    var = grad_and_vars[0][1]
    for g, v in grad_and_vars:
      assert v == var
      if g is not None:
        grads.append(g)
    if grads:
      if len(grads) > 1:
        sum_grad = tf.add_n(grads, name=var.op.name + '/sum_grads')
        sum_grad = tf.div(sum_grad, 1.0 * len(grads))
      else:
        sum_grad = grads[0]
      sum_grads.append((sum_grad, var))
  return sum_grads


def _add_gradients_summaries(grads_and_vars):
  """Add histogram summaries to gradients.

  Note: The summaries are also added to the SUMMARIES collection.

  Args:
    grads_and_vars: A list of gradient to variable pairs (tuples).

  Returns:
    The _list_ of the added summaries for grads_and_vars.
  """
  summaries = []
  for grad, var in grads_and_vars:
    if grad is not None:
      if isinstance(grad, tf.IndexedSlices):
        grad_values = grad.values
      else:
        grad_values = grad
      summaries.append(tf.summary.histogram(var.op.name + ':gradient',
                                            grad_values))
      summaries.append(tf.summary.histogram(var.op.name + ':gradient_norm',
                                            tf.global_norm([grad_values])))
    else:
      tf.logging.info('Var %s has no gradient', var.op.name)
  return summaries


class DeploymentConfig(object):
  """Configuration for deploying a model with `deploy()`.

  You can pass an instance of this class to `deploy()` to specify exactly
  how to deploy the model to build.  If you do not pass one, an instance built
  from the default deployment_hparams will be used.
  """

  def __init__(self,
               num_clones=1,
               devices=[],
               clone_on_cpu=False,
               replica_id=0,
               num_replicas=1,
               num_ps_tasks=0,
               worker_job_name='worker',
               ps_job_name='ps',
               clone_id_map={}):
    """Create a DeploymentConfig.

    The config describes how to deploy a model across multiple clones and
    replicas.  The model will be replicated `num_clones` times in each replica.
    If `clone_on_cpu` is True, each clone will placed on CPU.

    If `num_replicas` is 1, the model is deployed via a single process.  In that
    case `worker_device`, `num_ps_tasks`, and `ps_device` are ignored.

    If `num_replicas` is greater than 1, then `worker_device` and `ps_device`
    must specify TensorFlow devices for the `worker` and `ps` jobs and
    `num_ps_tasks` must be positive.

    Args:
      num_clones: Number of model clones to deploy in each replica.
      clone_on_cpu: If True clones would be placed on CPU.
      replica_id: Integer.  Index of the replica for which the model is
        deployed.  Usually 0 for the chief replica.
      num_replicas: Number of replicas to use.
      num_ps_tasks: Number of tasks for the `ps` job. 0 to not use replicas.
      worker_job_name: A name for the worker job.
      ps_job_name: A name for the parameter server job.

    Raises:
      ValueError: If the arguments are invalid.
    """
    if num_replicas > 1:
      if num_ps_tasks < 1:
        raise ValueError('When using replicas num_ps_tasks must be positive')
    if num_replicas > 1 or num_ps_tasks > 0:
      if not worker_job_name:
        raise ValueError('Must specify worker_job_name when using replicas')
      if not ps_job_name:
        raise ValueError('Must specify ps_job_name when using parameter server')
    if replica_id >= num_replicas:
      raise ValueError('replica_id must be less than num_replicas')
    self._num_clones = num_clones
    self._clone_on_cpu = clone_on_cpu
    self._replica_id = replica_id
    self._num_replicas = num_replicas
    self._num_ps_tasks = num_ps_tasks
    self._ps_device = '/job:' + ps_job_name if num_ps_tasks > 0 else ''
    self._worker_device = '/job:' + worker_job_name if num_ps_tasks > 0 else ''
    self._devices = devices
    self._clone_id_map = clone_id_map

  @property
  def num_clones(self):
    return self._num_clones

  @num_clones.setter
  def num_clones(self, n):
    self._num_clones = n

  @property
  def clone_id_map(self):
    return self._clone_id_map

  @property
  def devices(self):
    return self._devices

  @property
  def clone_on_cpu(self):
    return self._clone_on_cpu

  @property
  def replica_id(self):
    return self._replica_id

  @property
  def num_replicas(self):
    return self._num_replicas

  @property
  def num_ps_tasks(self):
    return self._num_ps_tasks

  @property
  def ps_device(self):
    return self._ps_device

  @property
  def worker_device(self):
    return self._worker_device

  def caching_device(self):
    """Returns the device to use for caching variables.

    Variables are cached on the worker CPU when using replicas.

    Returns:
      A device string or None if the variables do not need to be cached.
    """
    if self._num_ps_tasks > 0:
      return lambda op: op.device
    else:
      return None

  def clone_device(self, clone_index):
    """Device used to create the clone and all the ops inside the clone.

    Args:
      clone_index: Int, representing the clone_index.

    Returns:
      A value suitable for `tf.device()`.

    Raises:
      ValueError: if `clone_index` is greater or equal to the number of clones".
    """
    if clone_index >= self._num_clones:
      raise ValueError('clone_index must be less than num_clones')
    device = ''
    if self._num_ps_tasks > 0:
      device += self._worker_device
    if self._clone_on_cpu:
      device += '/device:CPU:0'
    else:
      if self._num_clones >= 1:
        clone_index = self._devices[clone_index] if len(self._devices) > 0 else clone_index
        device += '/device:GPU:%d' % clone_index
    return device

  def clone_scope(self, clone_index):
    """Name scope to create the clone.

    Args:
      clone_index: Int, representing the clone_index.

    Returns:
      A name_scope suitable for `tf.name_scope()`.

    Raises:
      ValueError: if `clone_index` is greater or equal to the number of clones".
    """
    if clone_index >= self._num_clones:
      raise ValueError('clone_index must be less than num_clones')
    scope = ''
    if self._num_clones > 1:
      scope = 'clone_%d' % clone_index
    return scope

  def optimizer_device(self):
    """Device to use with the optimizer.

    Returns:
      A value suitable for `tf.device()`.
    """
    if self._num_ps_tasks > 0 or self._num_clones > 0:
      return self._worker_device + '/device:CPU:0'
    else:
      return ''

  def inputs_device(self):
    """Device to use to build the inputs.

    Returns:
      A value suitable for `tf.device()`.
    """
    device = ''
    if self._num_ps_tasks > 0:
      device += self._worker_device
    device += '/device:CPU:0'
    return device

  def variables_device(self):
    """Returns the device to use for variables created inside the clone.

    Returns:
      A value suitable for `tf.device()`.
    """
    device = ''
    if self._num_ps_tasks > 0:
      device += self._ps_device
    device += '/device:CPU:0'

    class _PSDeviceChooser(object):
      """Slim device chooser for variables when using PS."""

      def __init__(self, device, tasks):
        self._device = device
        self._tasks = tasks
        self._task = 0

      def choose(self, op):
        if op.device:
          return op.device
        node_def = op if isinstance(op, tf.NodeDef) else op.node_def
        if node_def.op == 'Variable':
          t = self._task
          self._task = (self._task + 1) % self._tasks
          d = '%s/task:%d' % (self._device, t)
          return d
        else:
          return op.device

    if not self._num_ps_tasks:
      return device
    else:
      chooser = _PSDeviceChooser(device, self._num_ps_tasks)
      return chooser.choose



# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:26 上午
# @File    : download.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.utils import logger
from bs4 import BeautifulSoup
import os
import sys
import cv2
import numpy as nps
import threading
import requests
import re
import uuid
import os
try:
    import queue
except:
    import Queue as queue


def baidu_download(keyward, download_params, save_dir, process_queue=None):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='

    def __baidu_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url + str(t)
            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('"objURL":"(.*?)",', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                    logger.info("Downloading(%d) %s."%(num+1, str(each)))
                    try:
                        if each is not None:
                            pic = requests.get(each, timeout=7)
                        else:
                            continue
                    except BaseException:
                        logger.error("Couldnt download %s."%each)
                        continue
                    else:
                        # 分配唯一文件标识
                        file_folder = os.path.join(dir, 'test')
                        if not os.path.exists(file_folder):
                            os.makedirs(file_folder)
                        
                        file_path = os.path.join(file_folder, 'baidu_%s.jpg'%str(uuid.uuid4()))
                        with open(file_path, 'wb') as fp:
                            fp.write(pic.content)
                        num += 1

                        logger.info("Finish download %s ."% str(each))
                        # 加入等待处理队列
                        if waiting_process_queue is not None:
                            waiting_process_queue.put(file_path)

        # 添加结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if process_queue is not None:
        t = threading.Thread(target=__baidu_find_and_download, args=(process_queue, url, A, save_dir))
        t.start()
    else:
        __baidu_find_and_download(process_queue, url, A, save_dir)


def bing_download(keyward, download_params, save_dir, process_queue=None):
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    A = requests.Session()
    A.headers = headers
    #url = 'https://image.baidu.com/search/flip?tn=baiduimage&ie=utf-8&word=' + keyward + '&pn='
    url = 'http://cn.bing.com/images/async?q='+keyward+'&first={0}&count=35&relp=35&lostate=r&mmasync=1&dgState=x*175_y*848_h*199_c*1_i*106_r*0'

    def __bing_find_and_download(waiting_process_queue, search_url, session, dir, max_page_num=50):
        t = 0
        num = 0
        while t < max_page_num:
            Url = search_url.format(t*35+1)

            t = t+1
            try:
                Result = session.get(Url, timeout=7, allow_redirects=False)
            except BaseException:
                t = t + 60
                continue
            else:
                pic_url = re.findall('src="(.*?)"', Result.text, re.S)  # 先利用正则表达式找到图片url
                for each in pic_url:
                    logger.info("Downloading(%d) %s."%(num+1, str(each)))
                    try:
                        if each is not None:
                            pic = requests.get(each, timeout=7)
                        else:
                            continue
                    except BaseException:
                        logger.error("Couldnt download %s."%each)
                        continue
                    else:
                        # 分配唯一文件标识
                        file_folder = os.path.join(dir, 'test')
                        if not os.path.exists(file_folder):
                            os.makedirs(file_folder)
                        
                        file_path = os.path.join(file_folder, 'bing_%s.jpg'%str(uuid.uuid4()))
                        with open(file_path, 'wb') as fp:
                            fp.write(pic.content)
                        num += 1
                        logger.info("Finish download %s ."% str(each))

                        # 加入等待处理队列
                        if waiting_process_queue is not None:
                            waiting_process_queue.put(file_path)
        
        # 结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if process_queue is not None:
        t = threading.Thread(target=__bing_find_and_download, args=(process_queue, url, A, save_dir))
        t.start()
    else:
        __bing_find_and_download(process_queue, url, A, save_dir)


def google_download(keyward, download_params, save_dir, process_queue=None):
    pass


def vcg_download(keyword, download_params, download_save_dir, process_queue=None):
    # 视觉中国
    headers = {
        'Accept-Language': 'zh-CN,zh;q=0.8,zh-TW;q=0.7,zh-HK;q=0.5,en-US;q=0.3,en;q=0.2',
        'Connection': 'keep-alive',
        'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Upgrade-Insecure-Requests': '1'
    }
    download_data_type = 'image'
    if download_params is not None:
        download_data_type = download_params.get('download_data_type', 'image')
        if download_data_type not in ['image','video']:
            logger.warn('Only support donwload data_type: image or video')
            download_data_type = 'image'

    def __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, count):
        try:
            logger.info("Downloading(%d) %s."%(count+1, img_url))
            pic = requests.get(img_url, timeout=7)
        except BaseException:
            logger.error("Couldnt download %s."%img_url)
            return
        else:
            file_prefix = 'VCG_'+keyword + '_' + str(count)
            file_name = file_prefix + '.jpg' if download_data_type == 'image' else file_prefix + '.mp4'
            file_path = os.path.join(save_dir,  file_name)
            fp = open(file_path, 'wb')
            fp.write(pic.content)
            fp.close()
            logger.info("Finish download %s ."%img_url)
            
            if waiting_process_queue is not None:
                waiting_process_queue.put(file_path)

    def __vcg_find_and_download(waiting_process_queue, save_dir):
        A = requests.Session()
        A.headers = headers
        query_url = 'https://www.vcg.com/creative-image/{}'.format(keyword)
        if download_data_type == 'video':
            query_url = 'https://www.vcg.com/creative-video-search/{}'.format(keyword)

        download_count = 0
        content = None
        nav_page_list = []
        Result = A.get(query_url, timeout=7, allow_redirects=True)
        if Result.status_code != 200:
            logger.error("%s couldnt connect."%query_url)
            return

        content = Result.text
        soup = BeautifulSoup(content)
        # 1.step 分析待检索页面
        if download_data_type == 'image':
            pages = soup.findAll("a", class_="paginationClickBox")
            last_page = pages[-1]
            try:
                total_page_num = int(last_page.text)
            except:
                logger.error("Fail to parse nav page, use default 2.")
                total_page_num = 2

            page_nav_url = last_page.get('href').split('?')[0]
            for i in range(2, total_page_num+1):
                nav_page_list.append('%s?page=%d'%(page_nav_url, i))
        else:
            pages = soup.findAll("a", class_="_2IlL4")
            last_page = pages[-2]
            try:
                total_page_num = int(last_page.text)
            except:
                logger.error("Fail to parse nav page, use default 2.")
                total_page_num = 2

            page_nav_url = last_page.get('href').split('?')[0]
            for i in range(2, total_page_num+1):
                nav_page_list.append('%s?page=%d'%(page_nav_url, i))

        # 2.step 分析当前页面的图像
        img_url_list = []
        if download_data_type == 'image':
            # 分析图像资源
            img_list = soup.findAll("img", class_="lazyload_hk")
            img_url_list = ["http://{}".format(p.get("data-src")[2:]) for p in img_list]
        else:
            # 分析视频资源
            img_list = soup.findAll('source', type="image/webp")
            img_url_list = []
            video_name_list = []
            for p in img_list:
                kk = p.get('data-srcset')[2:].split('/')[3].split('?')[0].split('_')[0]
                if kk not in video_name_list:
                    video_name_list.append(kk)
                    img_url_list.append('http://gossv.cfp.cn/videos/mts_videos/medium/temp/{}.mp4'.format(kk))

        for img_url in img_url_list:
            __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, download_count)
            download_count += 1

        # 3.step 继续分析所有等待导航页面
        for page_url in nav_page_list:
            Result = A.get(page_url, timeout=7, allow_redirects=True)
            if Result.status_code != 200:
                continue

            content = Result.text
            soup = BeautifulSoup(content)

            img_url_list = []
            if download_data_type == 'image':
                # 分析图像资源
                img_list = soup.findAll("img", class_="lazyload_hk")
                img_url_list = ["http://{}".format(p.get("data-src")[2:]) for p in img_list]
            else:
                # 分析视频资源
                img_list = soup.findAll('source', type="image/webp")
                img_url_list = []
                video_name_list = []
                for p in img_list:
                    kk = p.get('data-srcset')[2:].split('/')[3].split('?')[0].split('_')[0]
                    if kk not in video_name_list:
                        video_name_list.append(kk)
                        img_url_list.append('http://gossv.cfp.cn/videos/mts_videos/medium/temp/{}.mp4'.format(kk))

            for img_url in img_url_list:
                __vcg_img_download(waiting_process_queue, save_dir, img_url, keyword, download_count)
                download_count += 1

        # 添加结束标记
        if waiting_process_queue is not None:
            waiting_process_queue.put(None)

    # 搜索和下载
    if not os.path.exists(os.path.join(download_save_dir, 'test')):
        os.makedirs(os.path.join(download_save_dir, 'test'))

    if process_queue is not None:
        t = threading.Thread(target=__vcg_find_and_download, args=(process_queue, os.path.join(download_save_dir, 'test')))
        t.start()
    else:
        __vcg_find_and_download(process_queue, os.path.join(download_save_dir, 'test'))





# -*- coding: UTF-8 -*-
# @Time    : 18-3-19
# @File    : normalized_error.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function




# -*- coding: UTF-8 -*-
# @Time    : 2020-06-25 23:29
# @File    : browser.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.crowdsource.browser_server import *
from antgo.dataflow.dataset.queue_dataset import *
from antgo.dataflow.recorder import *
from antgo.dataflow.dataset.parallel_read import MultiprocessReader
from antvis.client.httprpc import *
from antgo.dataflow.dataset.parallel_read import *
import requests
import json
from jinja2 import Environment, FileSystemLoader
import traceback

class BrowserDataRecorder(object):
  def __init__(self, maxsize=30):
    self.queue = queue.Queue()  # 不设置队列最大缓冲
    self.prepare_queue = queue.Queue(maxsize=maxsize)
    self.dump_dir = ''
    self.dataset_flag = 'TRAIN'
    self.dataset_size = 0
    self.tag_dir = ''

    # 5个线程，等待处理
    for _ in range(5):
      t = threading.Thread(target=self.asyn_record)
      t.daemon = True
      t.start()

  def _transfer_image(self, data):
    try:
      if len(data.shape) == 2:
        if data.dtype == np.uint8:
          transfer_result = data
        else:
          data_min = np.min(data)
          data_max = np.max(data)
          transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)
      else:
        assert (data.shape[2] == 3)
        transfer_result = data.astype(np.uint8)

      # save path
      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

      file_name = '%s.png' % str(uuid.uuid4())
      image_path = os.path.join(self.dump_dir, file_name)
      imwrite(image_path, transfer_result)
      return file_name
    except:
      logger.error('Couldnt transfer image data.')
      return None

  def asyn_record(self):
    while True:
      val = self.prepare_queue.get()
      if val is None:
        break

      # 加入到队列中
      assert(type(val) == dict)

      # 重新组织数据格式
      data = {}
      id = None
      for key, value in val.items():
        data_name = key
        if data_name not in data:
          data[data_name] = {}

        if 'data' in value or 'DATA' in value:
          if 'data' in value:
            data[data_name]['data'] = value['data']
            if data_name == 'id' or data_name == 'ID':
              id = value['data']
          else:
            data[data_name]['data'] = value['DATA']
            if data_name == 'id' or data_name == 'ID':
              id = value['data']

        if 'type' in value or 'TYPE' in value:
          if 'type' in value:
            data[data_name]['type'] = value['type']
          else:
            data[data_name]['type'] = value['TYPE']

        # if 'tag' in value or 'TAG' in value:
        #   if 'tag' in value:
        #     data[data_name]['tag'] = value['tag']
        #   else:
        #     data[data_name]['tag'] = value['TAG']

        # if 'id' in value or 'ID' in value:
        #   if 'id' in value:
        #     data[data_name]['id'] = value['id']
        #   else:
        #     data[data_name]['id'] = value['ID']

        data[data_name]['title'] = data_name

      # 转换数据到适合web
      web_data = []
      for name, body in data.items():
        if 'type' in body:
          if body['type'].upper() == 'IMAGE':
            body['width'] = body['data'].shape[1]
            body['height'] = body['data'].shape[0]
            body['data'] = '/static/data/%s'%self._transfer_image(body['data'])

        if 'type' not in body:
          body['type'] = 'STRING'

        if 'tag' not in body:
          body['tag'] = []

        body['dataset_flag'] = self.dataset_flag
        body['dataset_size'] = self.dataset_size
        web_data.append(body)

      # 如果从指定实验加载，则找寻是否存在以筛选标记
      if id is None:
        continue

      if os.path.exists(os.path.join(self.tag_dir, self.dataset_flag, '%s.json' % str(id))):
        # 加载json文件
        logger.info('Load record json from %s/%s'%(self.dataset_flag, '%s.json.' % str(id)))
        with open(os.path.join(self.tag_dir, self.dataset_flag, '%s.json' % str(id)), 'r') as fp:
          annotation = json.load(fp)
          # 确保标注数量和数据数量是一致的
          assert (len(web_data) == len(annotation))

          for a in web_data:
            is_update = False
            for b in annotation:
              # 检测数据和标注是否对应
              is_consistent = b['title'] == a['title']
              if 'id' in b:
                is_consistent = is_consistent and (a['id'] == b['id'])

                if not (a['id'] == b['id']):
                  logger.warn('Please whether data and record are consistent.')

              if is_consistent:
                a['tag'] = b['tag']
                is_update = True
                break

            if not is_update:
              if 'id' in a:
                logger.info('Annotation and data not consistence for %s.'%a['id'])
              else:
                logger.info('Annotation and data not consistence for %s.' % a['data'])

      # 加入队列，如果队列满，将阻塞
      self.queue.put(web_data)

  def record(self, val):
    self.prepare_queue.put(val)
    

class AntBrowser(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               ant_token,
               ant_host_ip,
               ant_host_port,
               ant_data_folder,
               ant_dataset,
               ant_dump_dir,
               **kwargs):
    super(AntBrowser, self).__init__(ant_name, ant_context, ant_token, **kwargs)
    self.ant_data_source = ant_data_folder
    self.dataset_name = ant_dataset
    self.dump_dir = ant_dump_dir
    self.host_ip = ant_host_ip
    self.host_port = ant_host_port
    self.rpc = None

  def start(self):
    # 1.step 获得数据集解析
    running_ant_task = None
    if self.token is not None:
      # 1.1.step load challenge task
      response = mlogger.getEnv().dashboard.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        # invalid token
        logger.error('couldnt load challenge task')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('prohibit submit challenge task frequently')
        exit(-1)
      elif response['status'] == 'OK':
        content = response['content']
        # maybe user token or task token
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('couldnt load challenge task')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('unknow error')
        exit(-1)

    if running_ant_task is not None:
      self.dataset_name = running_ant_task.dataset_name

    parse_flag = ''
    dataset_cls = AntDatasetFactory.dataset(self.dataset_name, parse_flag)

    if dataset_cls is None:
      logger.error('Couldnt find dataset parse class.')
      return

    # 2.step 配置记录器
    self.context.recorder = BrowserDataRecorder()

    # 3.step 启动浏览web服务
    browser_params = getattr(self.context.params, 'browser', None)
    tags = []
    if browser_params is not None:
      tags = browser_params.get('tags', [])

    # 3.1.step 准备web服务资源
    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    browser_static_dir = os.path.join(self.dump_dir, 'browser')
    if os.path.exists(browser_static_dir):
      shutil.rmtree(browser_static_dir)

    shutil.copytree(os.path.join(static_folder, 'resource', 'browser'), browser_static_dir)

    # 3.2.step 准备有效端口
    self.host_port = _pick_idle_port(self.host_port)
    self.rpc = HttpRpc("v1", "browser-api", "127.0.0.1", self.host_port)

    # base_url = '{}:{}'.format(self.host_ip, self.host_port)
    #
    # # 3.3.step 准备web配置文件
    # template_file_folder = os.path.join(static_folder, 'resource', 'browser', 'static')
    # file_loader = FileSystemLoader(template_file_folder)
    # env = Environment(loader=file_loader)
    # template = env.get_template('config.json')
    # output = template.render(BASE_URL=base_url)
    #
    # with open(os.path.join(browser_static_dir, 'static', 'config.json'), 'w') as fp:
    #   fp.write(output)

    # 3.3.step 状态
    train_offset, val_offset, test_offset = 0, 0, 0
    offset_configs = [{
      'dataset_flag': 'TRAIN',
      'dataset_offset': train_offset
    },{
      'dataset_flag': 'VAL',
      'dataset_offset': val_offset
    },{
      'dataset_flag': 'TEST',
      'dataset_offset': test_offset
    }]

    if self.context.params.browser is not None and 'offset' in self.context.params.browser:
      if 'TRAIN' in self.context.params.browser['offset']:
        train_offset = self.context.params.browser['offset']['TRAIN']
        offset_config = {
          'dataset_flag': 'TRAIN',
          'dataset_offset': train_offset
        }
        offset_configs[0] = offset_config

    if self.context.params.browser is not None and 'offset' in self.context.params.browser:
      if 'VAL' in self.context.params.browser['offset']:
        val_offset = self.context.params.browser['offset']['VAL']
        offset_config = {
          'dataset_flag': 'VAL',
          'dataset_offset': val_offset
        }
        offset_configs[1] = offset_config

    if self.context.params.browser is not None and 'offset' in self.context.params.browser:
      if 'TEST' in self.context.params.browser['offset']:
        test_offset = self.context.params.browser['offset']['TEST']
        offset_config = {
          'dataset_flag': 'TEST',
          'dataset_offset': test_offset
        }
        offset_configs[2] = offset_config

    dataset_flag = 'train'
    if self.context.params.browser is not None and 'dataset_flag' in self.context.params.browser:
      if self.context.params.browser['dataset_flag'].lower() in ['train', 'val', 'test']:
        dataset_flag = self.context.params.browser['dataset_flag'].lower()

    train_dataset = dataset_cls(dataset_flag, os.path.join(self.ant_data_source, self.dataset_name))
    self.context.recorder.dataset_flag = dataset_flag.upper()
    self.context.recorder.dataset_size = train_dataset.size
    self.context.recorder.dump_dir = os.path.join(self.dump_dir, 'browser', 'static', 'data')
    self.context.recorder.tag_dir = os.path.join(self.dump_dir, 'record')

    sample_offset = train_offset
    if dataset_flag == 'test':
      sample_offset = test_offset
    elif dataset_flag == 'val':
      sample_offset = val_offset

    if train_dataset.size == 0:
      logger.warn("Dont have waiting browser dataset(%s)."%dataset_flag)
      return

    logger.info('Browser %s dataset.' % dataset_flag)

    # 设置数据基本信息
    profile_config = {
      'dataset_flag': dataset_flag.upper(),
      'samples_num': train_dataset.size,
      'samples_num_checked': sample_offset
    }

    # 设置记录器偏移
    self.context.recorder.sample_index = sample_offset

    # 3.3.step 在线程中启动数据处理
    def _run_datagenerator_process():
      try:
        count = 0
        for data in self.context.data_generator(train_dataset):
          logger.info('Record data %d for browser.' % count)
          self.context.recorder.record(data)
          count += 1
      except:
        logger.info('Finish all records in browser %s dataset.' % dataset_flag)

    process = threading.Thread(target=_run_datagenerator_process)
    process.daemon = True
    process.start()

    # 3.4.step 启动web服务
    browser_mode = 'browser'
    if self.context.params.browser is not None and \
        'mode' in self.context.params.browser:
      browser_mode = self.context.params.browser['mode']

    browser_server_start(os.path.join(self.ant_data_source, self.dataset_name),
                         self.dump_dir,
                         self.context.recorder.queue,
                         tags,
                         self.host_port,
                         offset_configs,
                         profile_config,
                         browser_mode)



# -*- coding: UTF-8 -*-
# @Time    : 2019/1/22 1:16 PM
# @File    : activelearning.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.resource.html import *
from antgo.ant.base import *
from antgo.activelearning.samplingmethods.kcenter_greedy import *
from antgo.crowdsource.activelearning_server import *
from antgo.dataflow.common import *
from antgo.dataflow.recorder import *
from antvis.client.httprpc import *
from multiprocessing import Process, Queue
from antgo.task.task import *
from scipy.stats import entropy
import traceback
import subprocess
import os
import socket
import requests
import json
import zipfile

def _is_open(check_ip, port):
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  try:
    s.connect((check_ip, int(port)))
    s.shutdown(2)
    return True
  except:
    return False


def _pick_idle_port(from_port=40000, check_count=100):
  check_port = from_port
  while check_count:
    if not _is_open('127.0.0.1', check_port):
      break

    logger.warn('Port %d is occupied, try to use %d port.'%(int(check_port), int(check_port + 1)))

    check_port += 1
    check_count -= 1

    if check_count == 0:
      check_port = None

  if check_port is None:
    logger.warn('Couldnt find valid free port.')
    exit(-1)

  return check_port


class AntActiveLearning(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               ant_data_folder,
               ant_dump_dir,
               ant_token,
               ant_task_config=None,
               **kwargs):
    super(AntActiveLearning, self).__init__(ant_name, ant_context, ant_token, **kwargs)

    self.skip_first_training = self.context.params.system['skip_training']
    self.max_iterators = self.context.params.activelearning.get('max_iterators', 10)
    if self.max_iterators is None:
      self.max_iterators = 10

    self.dump_dir = ant_dump_dir
    self.web_server_port = self.context.params.system['port']
    self.web_server_port = int(self.web_server_port) if self.web_server_port is not None else None
    self.html_template = kwargs.get('html_template', None)
    self.keywords_template = {}

    self.ant_task_config = ant_task_config
    self.ant_data_source = ant_data_folder
    self.devices = self.context.params.system['devices']
    self.rpc = None

  def _core_set_algorithm(self, unlabeled_pool, num):
    if num > len(unlabeled_pool):
      logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
      return unlabeled_pool

    # data: 1XC (C表示特征维度)
    feature_data = np.array([data['feature'] for data in unlabeled_pool])
    channels = feature_data.shape[-1]
    kcg = kCenterGreedy(feature_data.reshape(-1, channels))
    next_selected = kcg.select_batch(model=None, already_selected=[], N=num)
    next_selected = [unlabeled_pool[int(s)] for s in next_selected]

    return next_selected

  def _entroy_algorithm(self, unlabeled_pool, num):
    if num > len(unlabeled_pool):
      logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
      return unlabeled_pool

    unlabeled_entropy = []
    for index, data in enumerate(unlabeled_pool):
      # data: DxC（D表示实验次数，C表示概率信息）
      assert(len(data['feature'].shape) == 2)
      p = np.mean(data['feature'], axis=0)
      unlabeled_entropy.append((entropy(p), index))

    ordered_unlabeled = sorted(unlabeled_entropy, key=lambda x: x[0], reverse=True)
    next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
    return next_selected

  def _variance_algorithm(self, unlabeled_pool, num):
    if num > len(unlabeled_pool):
      logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))      
      return unlabeled_pool

    unlabeled_var = []
    for index, data in enumerate(unlabeled_pool):
      # data: DxC (D表示试验次数，C表示概率信息)
      assert(len(data['feature'].shape) == 2)
      var_val = np.var(data['feature'], axis=0)
      unlabeled_var.append(((float)(np.mean(var_val)), index))
    
    ordered_unlabeled = sorted(unlabeled_var, key=lambda x: x[0], reverse=True)
    next_selected = [unlabeled_pool[s[1]] for s in ordered_unlabeled[0:num]]
    return next_selected

  def _unform_sampling_algorithm(self, unlabeled_pool, num):
    if num > len(unlabeled_pool):
      logger.warn("Dont have enough unlabeled data(%d), return all."%len(unlabeled_pool))
      return unlabeled_pool
    
    return np.random.choice(unlabeled_pool, num, False)

  def _waiting_label_sample_select(self, unlabeled_pool, num):
    sampling_strategy = self.context.params.activelearning.get('sampling_strategy', 'entropy')
    logger.info('Using %s Strategy to select waiting label data.'%sampling_strategy)
    if sampling_strategy == 'coreset':
      return self._core_set_algorithm(unlabeled_pool, num)
    elif sampling_strategy == 'entropy':
      return self._entroy_algorithm(unlabeled_pool, num)
    elif sampling_strategy == 'variance':
      return self._variance_algorithm(unlabeled_pool, num)
    elif sampling_strategy == 'random':
      return self._unform_sampling_algorithm(unlabeled_pool, num)
    else:
      return self._unform_sampling_algorithm(unlabeled_pool, num)

  def start(self):
    # 0.step loading challenge task
    running_ant_task = None
    if self.token is not None:
      # 0.step load challenge task
      response = self.context.dashboard.challenge.get(command=type(self).__name__)
      if response['status'] is None:
        # invalid token
        logger.error('Couldnt load challenge task.')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('Prohibit submit challenge task frequently.')
        exit(-1)
      elif response['status'] == 'OK':
        # maybe user token or task token
        content = response['content']
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('Couldnt load challenge task.')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('Unknow error.')
        exit(-1)

    if running_ant_task is None:
      # 0.step load custom task
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('Couldnt load custom task.')
        exit(0)
      running_ant_task = custom_task

    assert (running_ant_task is not None)

    # 配置html模板信息
    self.keywords_template['TASK_TITLE'] = running_ant_task.task_name
    self.keywords_template['TASK_TYPE'] = running_ant_task.task_type
    self.keywords_template['UNLABELED_NUM'] = 0
    self.keywords_template['LABEL_COMPLEX'] = 0

    # dataset
    dataset = \
        running_ant_task.dataset('train',os.path.join(self.ant_data_source, running_ant_task.dataset_name), running_ant_task.dataset_params)

    # prepare workspace
    if not os.path.exists(os.path.join(self.main_folder, 'web', 'static', 'data')):
      os.makedirs(os.path.join(self.main_folder, 'web', 'static', 'data'))

    annotation_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'annotations')
    if not os.path.exists(annotation_folder):
      os.makedirs(annotation_folder)

    data_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'images')
    if not os.path.exists(data_folder):
      os.makedirs(data_folder)

    download_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'download')
    if not os.path.exists(download_folder):
      os.makedirs(download_folder)

    upload_folder = os.path.join(self.main_folder, 'web', 'static', 'data', 'upload')
    if not os.path.exists(upload_folder):
      os.makedirs(upload_folder)

    # 数据队列
    request_queue = Queue()

    # launch web server
    if self.web_server_port is None:
      self.web_server_port = 10000
    self.web_server_port = _pick_idle_port(self.web_server_port)

    logger.info('Launch active learning web server on port %d.'%self.web_server_port)
    process = multiprocessing.Process(target=activelearning_web_server,
                                      args=('activelearning',
                                            self.main_folder,
                                            self.html_template,
                                            self.keywords_template,
                                            running_ant_task,
                                            self.web_server_port,
                                            os.getpid(),
                                            download_folder,
                                            upload_folder,
                                            request_queue))
    process.daemon = True
    process.start()
    logger.info('Waiting 5 seconds for launching web server.')
    time.sleep(5)

    self.rpc = HttpRpc('v1','activelearning','127.0.0.1',self.web_server_port)
    avg_analyze_time = 0

    # prepare waiting unlabeled data
    try_iter = 0
    experiment_id = None
    while try_iter < self.max_iterators:
      logger.info('Finding unlabeled and cnadidates size.')
      unlabeled_dataset_size = dataset.unlabeled_size()
      labeled_dataset_size = dataset.candidates_size()

      min_sampling_num = self.context.params.activelearning.get('min_sampling_num', None)
      if min_sampling_num is not None:
        min_sampling_num = (int)(min_sampling_num)
        if unlabeled_dataset_size < min_sampling_num:
          logger.info('Active learning is over. (unlabeled sample size < %d).'%min_sampling_num)
          return

      if unlabeled_dataset_size < 10:
        logger.info('Active learning is over. (unlabeled sample size < 10).')
        return

      logger.info("Round %d, unlabeled dataset size %d, labeled dataset size %d."%(try_iter, unlabeled_dataset_size, labeled_dataset_size))
      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

      # 当前阶段
      self.stage = "ACTIVELEARNING-TRAIN-ROUND-%d"%try_iter

      # 通知处理状态(开始未标注数据集挖掘)
      self.rpc.state.patch(round=try_iter, process_state="UNLABEL-PREPARE")
      # 开始分析时间
      analyze_start_time = time.time()
      if try_iter == 0:
        experiment_id = self.context.from_experiment

      if (not self.skip_first_training or try_iter > 0) and labeled_dataset_size > 0:
        # shell call
        logger.info('Start training using all labeled data (%d iter).'%try_iter)

        if os.path.exists(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter)):
          shutil.rmtree(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter))
        os.makedirs(os.path.join(self.dump_dir, 'try_round_train_%d'%try_iter))

        cmd_shell = 'antgo train --main_file=%s --main_param=%s' % (self.main_file, self.main_param)
        cmd_shell += ' --dump=%s/%s' % (self.dump_dir, 'try_round_train_%d'%try_iter)
        cmd_shell += ' --main_folder=%s' % self.main_folder
        cmd_shell += ' --task=%s' % self.ant_task_config.split('/')[-1]
        if experiment_id is not None:
          cmd_shell += ' --from_experiment=%s' % experiment_id
        cmd_shell += ' --candidate'
        cmd_shell += ' --devices=%s' % self.devices
        cmd_shell += ' --dataset=%s/train' % running_ant_task.dataset_name
        cmd_shell += ' --name=%s_train_round_%d' % (self.ant_name, try_iter)
        training_p = \
            subprocess.Popen('%s > %s.log' % (cmd_shell, '%s_try_rounnd_train_%d'%(self.name, try_iter)), 
                              shell=True, 
                              cwd=self.main_folder)

        # waiting untile finish training
        training_p.wait()

        # 根据返回结果，判断是否正常结束
        if training_p.returncode != 0:
          logger.error('Training process exit anomaly.')
          exit(-1)

        # 获取训练完成后的实验目录ß
        experiment_prefix = 'try_round_train_%d'%try_iter
        experiment_id = None
        for k in os.listdir(os.path.join(self.dump_dir, experiment_prefix)):
          if k[0] == '.':
            continue

          if os.path.isdir(os.path.join(self.dump_dir, experiment_prefix, k)):
            experiment_id = os.path.join(self.dump_dir, experiment_prefix, k)
            break
      
      # 2.step inference using unlabeled data
      # 当前阶段
      self.stage = "ACTIVELEARNING-ANALYZE-ROUND-%d"%try_iter
      if os.path.exists(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter)):
        shutil.rmtree(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter))
      os.makedirs(os.path.join(self.dump_dir, 'try_round_analyze_%d'%try_iter))        

      logger.info('Start analyze all unlabeled data distribution (%d iter).'%try_iter)
      cmd_shell = 'antgo predict --main_file=%s --main_param=%s'%(self.main_file, self.main_param)
      cmd_shell += ' --main_folder=%s' % self.main_folder
      cmd_shell += ' --dump=%s/%s' % (self.dump_dir, 'try_round_analyze_%d'%try_iter)
      if experiment_id is not None:
        cmd_shell += ' --from_experiment=%s' % experiment_id
      cmd_shell += ' --task_t=%s' % running_ant_task.task_type
      cmd_shell += ' --unlabel'
      cmd_shell += ' --devices=%s' % self.devices
      cmd_shell += ' --dataset=%s/train' % running_ant_task.dataset_name
      cmd_shell += ' --name=%s_predict_round_%d' % (self.ant_name, try_iter)
      inference_p = subprocess.Popen('%s > %s.log' %(cmd_shell, '%s_try_rounnd_analyze_%d'%(self.name, try_iter)), 
                                      shell=True, 
                                      cwd=self.main_folder)

      # waiting untile finish inference
      inference_p.wait()

      # 根据返回结果，判断是否正常结束
      if inference_p.returncode != 0:
        logger.error('Inference process exit anomaly.')
        exit(-1)

      # 获取推断完成后的实验目录
      inference_experiment_prefix = 'try_round_analyze_%d'%try_iter
      inference_experiment_id = ''
      for k in os.listdir(os.path.join(self.dump_dir, inference_experiment_prefix)):
        if k[0] == '.':
          continue

        if os.path.isdir(os.path.join(self.dump_dir, inference_experiment_prefix, k)):
          inference_experiment_id = k
          break
      
      record_reader = RecordReader(os.path.join(self.dump_dir, inference_experiment_prefix, inference_experiment_id, 'record'))
      unlabeled_pool = []
      for ss in record_reader.iterate_read('groundtruth', 'predict'):
        gt, feature = ss
        unlabeled_pool.append({'file_id': gt['file_id'], 'feature': feature, 'id': gt['id']})

      select_size = self.context.params.activelearning.get('min_sampling_num', None)
      if select_size is None:
        min_sampling_ratio = self.context.params.activelearning.get('min_sampling_ratio', None)
        if min_sampling_ratio is None:
          min_sampling_ratio = 0.1
        select_size = int(len(unlabeled_pool) * min_sampling_ratio)
        if select_size == 0:
          select_size = len(unlabeled_pool)

      if select_size == 0:
        logger.info('Active learning is over. (selecting size == 0.')
        return

      next_selected = self._waiting_label_sample_select(unlabeled_pool, select_size)
      if len(next_selected) == 0:
        logger.info("Active learning is over. (selecting size == 0).")
        return

      logger.info("Round %d, selecting size %d by %s method."%(try_iter, select_size, self.context.params.activelearning.get('sampling_strategy', 'entropy')))

      # 结束分析时间
      analyze_end_time = time.time()

      # 获得平均分析时间
      avg_analyze_time = (avg_analyze_time * try_iter + (analyze_end_time - analyze_start_time)) / (try_iter + 1)

      next_unlabeled_sample_ids = []
      for f in next_selected:
        next_unlabeled_sample_ids.append((f['file_id'], f['id']))

      # 打包等待下一步进行标注的样本
      tar_file = "round_%d.tar.gz"%try_iter
      tar_path = os.path.join(download_folder, tar_file)
      if os.path.exists(tar_path):
        os.remove(tar_path)

      tar = tarfile.open(tar_path, "w:gz")
      for next_unlabeled_sample_id, _ in next_unlabeled_sample_ids:
        tar.add(os.path.join(dataset.dir, next_unlabeled_sample_id),
                arcname="round_%d/%s"%(try_iter,next_unlabeled_sample_id))
      tar.close()

      # 研究模式，在研究模式下，标注结果自动获取
      if self.context.params.system['research']:
        # 获取标注数据，并打包保存
        logger.info("Research: auto get label data.")
        if not os.path.exists(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter)):
          os.makedirs(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter))

        for file_id, id in next_unlabeled_sample_ids:
          _, label = dataset.at(id, file_id)
          label.update({'file_id': file_id, 'id': id})
          
          # 自动生成子目录
          if '/' in file_id:
            if not os.path.exists(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id.split('/')[0])):
              os.makedirs(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id.split('/')[0]))

          # label 写成文件
          with open(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id), 'w') as fp:
            json.dump(label, fp)

        # 使用tar,打包
        logger.info("Research: warp label data.")
        tar = tarfile.open(os.path.join(self.dump_dir, 'try_round_auto_label_%d.tar.gz'%try_iter), "w:gz")
        for file_id, _ in next_unlabeled_sample_ids:
          tar.add(os.path.join(self.dump_dir, 'try_round_auto_label_%d'%try_iter, file_id),
                  arcname="try_round_auto_label_%d/%s"%(try_iter, file_id))
        tar.close()

        logger.info("Research: copy to %s."%upload_folder)
        shutil.copy(os.path.join(self.dump_dir, 'try_round_auto_label_%d.tar.gz'%try_iter), upload_folder)
        request_queue.put({'FILE': 'try_round_auto_label_%d.tar.gz'%try_iter, 'ROUND': try_iter})

      # 通知处理状态(已经准备好未标注数据集)
      self.rpc.state.patch(round=try_iter,
                           process_state="UNLABEL-RESET",
                           unlabel_dataset=tar_file,
                           unlabeled_size=unlabeled_dataset_size,
                           labeled_size=labeled_dataset_size,
                           round_size=len(next_selected))

      while True:
        logger.info('Waiting label (human in loop) in round %d.'%try_iter)
        request_content = request_queue.get()

        # 1.step 解压数据集文件
        logger.info("Untar label dataset.")
        request_label_dataset = request_content['FILE']
        request_round = request_content['ROUND']
        if request_round != try_iter:
          logger.error('Request label round not consistent.')
          self.rpc.state.patch(round=try_iter, process_state="LABEL-ERROR")
          continue

        # 2.step 建立文件夹
        if os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
          shutil.rmtree(os.path.join(upload_folder, "round_%d"%request_round))

        # uncompress
        if request_label_dataset.endswith('tar') or request_label_dataset.endswith('tar.gz'): 
          tar = tarfile.open(os.path.join(upload_folder, request_label_dataset))
          tar_list = tar.getnames()
          tar.extractall(upload_folder)
          tar.close()

          if len(tar_list) == 0:
            logger.error('Untar file error.')
            continue

          if not os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
            # 将解压后的文件夹，修改为 "round_%d"%request_round
            file_name = os.path.normpath(tar_list[0]).split('/')[0]
            shutil.move(os.path.join(upload_folder, file_name), os.path.join(upload_folder, "round_%d"%request_round))
        elif request_label_dataset.endswith('zip'):
          zip_file = zipfile.ZipFile(os.path.join(upload_folder, request_label_dataset))
          zip_list = zip_file.namelist()
          for f in zip_list:
              zip_file.extract(f, upload_folder)
          zip_file.close()

          if len(zip_list) == 0:
            logger.error('Unzip file error.')
            continue

          if not os.path.exists(os.path.join(upload_folder, "round_%d"%request_round)):
            # 将解压后的文件夹，修改为 "round_%d"%request_round
            file_name = os.path.normpath(zip_list[0]).split('/')[0]
            shutil.move(os.path.join(upload_folder, file_name), os.path.join(upload_folder, "round_%d"%request_round))
        else:
          logger.error('Dont support upload type.')
          continue
        
        # 3.step 检查数据集标准是否符合标准
        logger.info("Check labeled dataset format.")
        is_ok = dataset.check_candidate(next_unlabeled_sample_ids, os.path.join(upload_folder, "round_%d"%request_round))
        if not is_ok:
          logger.error("Dataset format maybe error, need to update label.")
          self.rpc.state.patch(round=try_iter, process_state="LABEL-ERROR")
          continue

        break

      self.rpc.state.patch(round=try_iter,
                           process_state='LABEL-FINISH',
                           next_round_waiting=avg_analyze_time)

      logger.info('Prepare candidates.txt by using new labeled data.')
      for sample_file, sample_id in next_unlabeled_sample_ids:
        if os.path.exists(os.path.join(upload_folder, 'round_%d'%try_iter, sample_file)):
          dataset.make_candidate(sample_id,
                                  sample_file,
                                  os.path.join(upload_folder, 'round_%d'%try_iter, sample_file),
                                  'OK')

      logger.info('Finish round %d label, start next round.'%try_iter)
      # increment round
      try_iter += 1



# -*- coding: UTF-8 -*-
# @Time    : 2020-06-25 23:30
# @File    : watch_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import os
import shutil
import signal
from antgo.crowdsource.base_server import *
from antgo.utils import logger
import traceback
import sys
import uuid
import json
import time
import tarfile


class IndexHandler(BaseHandler):
  @gen.coroutine
  def get(self, *args, **kwargs):
    self.response(RESPONSE_STATUS_CODE.SUCCESS, content={})


class WatchApiHandler(BaseHandler):
  @property
  def data_queue(self):
    return self.settings['data_queue']

  @gen.coroutine
  def post(self, *args, **kwargs):
    datasource_type = self.get_argument('datasource_type', '')
    datasource_address = self.get_argument('datasource_address', '')
    datasource_keywards = self.get_argument('datasource_keywards', '')

    if datasource_type not in ['spider']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED)
      return

    if datasource_address not in ['baidu', 'bing']:
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED)
      return

    self.db['watch'].push({
      'watch_time': time.time(),
      'datasource_type': datasource_type,
      'datasource_address': datasource_address,
      'datasource_keywards': datasource_keywards,
      'watch_id': len(self.db['watch']),
      'watch_data': []
    })

    self.data_queue.push([datasource_type, {'datasource_address': datasource_address, 'datasource_keywards': datasource_keywards}])


class UpdateApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    data_type = self.get_argument('processed_type', '')
    data_width = self.get_argument('processed_width', '')
    data_height = self.get_argument('processed_height', '')
    data = self.get_argument('processed_data', '')
    data_tag = self.get_argument('processed_tag', '')
    data_id = self.get_argument('processed_id', '')
    watch_id = self.get_argument('processed_watch_id', '')

    watch_id = (int)(watch_id)
    if watch_id >= len(self.db['watch']):
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED, message='watch id not in list')
      return

    self.db['watch'][watch_id]['watch_data'].push({
      'data_type': data_type,
      'data_width': data_width,
      'data_height': data_height,
      'data': data,
      'data_tag': data_tag,
      'data_id': data_id,
      'data_index': len(self.db['watch'][watch_id]['watch_data']),
      'data_del': False,
      'data_select': False
    })
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class DelApihandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    watch_id = self.get_argument('watch_id', '')
    watch_id = (int)(watch_id)
    if watch_id >= len(self.db['watch']):
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED, message='watch id not in list')
      return

    data_index = self.get_argument('data_index', '')
    data_index = (int)(data_index)
    if data_index >= len(self.db['watch'][watch_id]['watch_data']):
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED, message='data index not in list')
      return

    self.db['watch'][watch_id]['watch_data'][data_index]['data_del'] = True
    self.response(RESPONSE_STATUS_CODE.SUCCESS)

  @gen.coroutine
  def get(self, *args, **kwargs):
    del_data = []
    for watch_id in range(len(self.db['watch'])):
      for data in self.db['watch'][watch_id]['watch_data']:
        if data['data_del']:
          del_data.append(data)

    self.response(RESPONSE_STATUS_CODE.SUCCESS, content=del_data)


class SelectApiHandler(BaseHandler):
  @gen.coroutine
  def post(self, *args, **kwargs):
    watch_id = self.get_argument('watch_id', '')
    watch_id = (int)(watch_id)
    if watch_id >= len(self.db['watch']):
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED, message='watch id not in list')
      return

    data_index = self.get_argument('data_index', '')
    data_index = (int)(data_index)
    if data_index >= len(self.db['watch'][watch_id]['watch_data']):
      self.response(RESPONSE_STATUS_CODE.REQUEST_NOT_ACCEPTED, message='data index not in list')
      return

    self.db['watch'][watch_id]['watch_data'][data_index]['data_select'] = True
    self.response(RESPONSE_STATUS_CODE.SUCCESS)


class TarApiHandler(BaseHandler):
  tar_v = 1
  @gen.coroutine
  def post(self, *args, **kwargs):
    tar = None
    for watch_id in range(len(self.db['watch'])):
      for data in self.db['watch'][watch_id]['watch_data']:
        if data['data_select']:
          if tar is None:
            tar_path = os.path.join(self.static_path, 'v-%d.tar.gz'%TarApiHandler.tar_v)
            tar = tarfile.open(tar_path, 'w:gz')

          tar.add(data['data'])
        pass

    TarApiHandler.tar_v += 1
    pass


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def watch_server_start(data_path,
                         browser_dump_dir,
                         response_queue,
                         tags,
                         server_port,
                         offset_configs,
                         profile_config,
                         browser_mode='screening'):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)

  # 0.step define http server port
  define('port', default=server_port, help='run on port')

  try:
    # 静态资源目录
    browser_static_dir = os.path.join(browser_dump_dir, 'watch')
    # 数据数据目录
    browser_dump_dir = os.path.join(browser_dump_dir, 'record')
    if not os.path.exists(browser_dump_dir):
      os.makedirs(browser_dump_dir)

    # 2.step launch web server
    db = {'data': [], 'user': {}, 'dataset': {}}
    for offset_config in offset_configs:
      db['dataset'][offset_config['dataset_flag']] = {'offset':offset_config['dataset_offset']}

    db['dataset'][profile_config['dataset_flag']]['samples_num'] = profile_config['samples_num']
    db['dataset'][profile_config['dataset_flag']]['samples_num_checked'] = profile_config['samples_num_checked']
    db['state'] = profile_config['dataset_flag']

    settings = {
      'static_path': os.path.join(browser_static_dir, 'static'),
      'dump_path': browser_dump_dir,
      'port': server_port,
      'response_queue': response_queue,
      'tags': tags,
      'data_folder': data_path,
      'cookie_secret': str(uuid.uuid4()),
      'db': db,
      'browser_mode': browser_mode
    }

    app = tornado.web.Application(handlers=[(r"/browser-api/prev/", PrevApiHandler),
                                            (r"/browser-api/next/", NextApiHandler),
                                            (r"/browser-api/operator/", OperatorApiHandler),
                                            (r"/browser-api/entry/", EntryApiHandler),
                                            (r"/browser-api/file/", FileApiHandler),
                                            (r"/browser-api/download/", DownloadApiHandler),
                                            (r"/browser-api/config/", ConfigApiHandler),
                                            (r'/(.*)', tornado.web.StaticFileHandler,
                                             {"path": browser_static_dir, "default_filename": "index.html"}),],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(options.port)

    logger.info('demo is providing server on port %d' % server_port)
    tornado.ioloop.IOLoop.instance().start()
    logger.info('demo stop server')
  except GracefulExitException:
    logger.info('demo server exit')
    sys.exit(0)
  except KeyboardInterrupt:
    pass

# browser_server_start()


# -*- coding: UTF-8 -*-
# @Time : 29/03/2018
# @File : lip.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np

__all__ = ['LIP']
class LIP(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(LIP, self).__init__(train_or_test, dir, params)
    assert (train_or_test in ['train', 'sample', 'val'])

    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    self.data_list = []
    parse_file = os.path.join(self.dir, '%s' % ('train_id.txt' if self.train_or_test == 'train' else 'val_id.txt'))
    image_folder = os.path.join(self.dir, '%s_images' % ('train' if self.train_or_test == 'train' else 'val'))
    label_folder = os.path.join(self.dir, '%s_segmentations' % ('train' if self.train_or_test == 'train' else 'val'))
    with open(parse_file, 'r') as fp:
      content = fp.readline()
      while content:
        content = content.replace('\n', '')
        self.data_list.append((os.path.join(image_folder, content),
                               os.path.join(label_folder, content)))

        content = fp.readline()

    self.ids = list(range(len(self.data_list)))

    self.inv_category_map = {
      0: 'background',
      1: 'Hat',
      2: 'Hair',
      3: 'Sunglasses',
      4: 'Upper-clothes',
      5: 'Dress',
      6: 'Coat',
      7: 'Socks',
      8: 'Pants',
      9: 'Glove',
      10: 'Scarf',
      11: 'Skirt',
      12: 'Jumpsuits',
      13: 'Face',
      14: 'Right-arm',
      15: 'Left-arm',
      16: 'Right-leg',
      17: 'Left-leg',
      18: 'Right-shoe',
      19: 'Left-shoe'
    }
    self.category_map = {v: k for k, v in self.inv_category_map.items()}

  @property
  def size(self):
    return len(self.ids)

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1

      # idxs = np.arange(len(self.ids))
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)

      for k in idxs:
        image_file, label_file = self.data_list[k]
        image = imread('%s.jpg' % image_file)
        label = imread('%s.png' % label_file)
        if len(label.shape) > 2:
          label = label[:, :, 0]

        included_cls = getattr(self, 'included', None)
        if included_cls is not None:
          included_cls = [int(a) for a in included_cls.split(',')]

        excluded_cls = getattr(self, 'excluded', None)
        if excluded_cls is not None:
          excluded_cls = [int(a) for a in excluded_cls.split(',')]

        existed_cls = None
        if included_cls is not None or excluded_cls is not None:
          existed_cls = list(set(label.flatten().tolist()))
          existed_cls.sort()

        valid_objs_num = -1
        valid_objs_index = []
        if included_cls is not None:
          valid_objs_num = len(existed_cls)

          for cls_label in existed_cls:
            if cls_label not in included_cls:
              cls_position = np.where(label == cls_label)
              label[cls_position] = 0
              valid_objs_num -= 1
            else:
              valid_objs_index.append(cls_label)

        if excluded_cls is not None:
          for cls_label in existed_cls:
            if cls_label in excluded_cls:
              cls_position = np.where(label == cls_label)
              label[cls_position] = 0
              valid_objs_num -= 1
            else:
              valid_objs_index.append(cls_label)

        if valid_objs_num == 0:
          continue

        # split to channels for every object
        segmentation_objs = []
        height, width = image.shape[0:2]
        for obj_index in valid_objs_index:
          seg = np.zeros((height, width), dtype=np.uint8)
          seg[np.where(label == obj_index)] = 1

          segmentation_objs.append(seg)

        yield [image, {'segmentation': segmentation_objs, 'segmentation_map': label, 'id': k}]

  def at(self, id):
    image_file, label_file = self.data_list[id]
    image = imread(image_file)
    label = imread(label_file)
    return [image, {'segmentation': label, 'segmentation_map': label, 'id': id}]

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    val_dataset = LIP('val', self.dir)

    return self, val_dataset



# -*- coding: UTF-8 -*-
# @Time    : 17-12-27
# @File    : vggface.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset import *
import os
import numpy as np

VGGFACE_URL = 'http://www.robots.ox.ac.uk/~vgg/data/vgg_face/vgg_face_dataset.tar.gz'
class VGGFace(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(VGGFace, self).__init__(train_or_test, dir)
    assert(train_or_test in ['train', 'sample'])
    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    # 0.step maybe download
    if not os.path.exists(os.path.join(self.dir, 'vgg_face_dataset.tar.gz')):
      self.download(self.dir, file_names=[], default_url=VGGFACE_URL, auto_untar=True, is_gz=True)
    
    if not os.path.exists(os.path.join(self.dir, 'data')) or len(os.listdir(os.path.join(self.dir, 'data'))) < 500:
      if not os.path.exists(os.path.join(self.dir, 'data')):
        os.makedirs(os.path.join(self.dir, 'data'))
      # download all images from web
      for file in os.listdir(os.path.join(self.dir, 'vgg_face_dataset', 'files')):
        if file[0] == '.':
          continue
        
        with open(os.path.join(self.dir, 'vgg_face_dataset', 'files', file), 'r') as fp:
          content = fp.readline()
          while content:
            image_id, image_url, x_1, y_1, x_2, y_2, pose, _1, _2 =\
              [i for i in content.replace('\n', '').split(' ') if i != '']
            
            # download every image as image_id.jpg
            name = os.path.normpath(image_url).split('/')[-1]
            if not os.path.exists(os.path.join(self.dir, 'data', name)):
              try:
                download(image_url, os.path.join(self.dir, 'data'))
              except:
                logger.error('%s couldnt be downloaded'%image_url)
            else:
              logger.info('%s has been existed'%name)
            
            content = fp.readline()
    
    # 1.step data files
    self._persons_file = []
    self._persons_annotation = []
    self._persons_id_str = []
    self._persons_id = []
    for file in os.listdir(os.path.join(self.dir, 'vgg_face_dataset', 'files')):
      if file[0] == '.':
        continue
  
      with open(os.path.join(self.dir, 'vgg_face_dataset', 'files', file), 'r') as fp:
        content = fp.readline()
        while content:
          image_id, _url, x_1, y_1, x_2, y_2, pose, _1, _2 =\
            [i for i in content.replace('\n', '').split(' ') if i != '']
          image_name = os.path.normpath(_url).split('/')[-1]
          if os.path.exists(os.path.join(self.dir, 'data', image_name)):
            self._persons_file.append(os.path.join(self.dir, 'data', image_name))
            bbox = np.zeros((1, 4))
            bbox[0, 0] = float(x_1)
            bbox[0, 1] = float(y_1)
            bbox[0, 2] = float(x_2)
            bbox[0, 3] = float(y_2)
            self._persons_annotation.append({'bbox': bbox, 'pose': float(pose), 'category': file.split('.')[0]})
            self._persons_id_str.append(file)
          
          content = fp.readline()

    id_set = set(self._persons_id_str)
    person_id_map = {}
    for s_i, s in enumerate(id_set):
      person_id_map[s] = s_i

    for person_id_str in self._persons_id_str:
      person_id = person_id_map[person_id_str]
      self._persons_id.append(person_id)

    self.ids = list(range(len(self._persons_file)))

    # fixed seed
    self.seed = time.time()
  
  @property
  def size(self):
    return len(self.ids)
  
  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1
    
      # idxs = np.arange(len(self.ids))
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)
        
      for k in idxs:
        person_file = self._persons_file[k]
        person_image = imread(person_file)
        person_annotation = self._persons_annotation[k]
        person_id = self._persons_id[k]
        person_annotation.update({'category_id': person_id,
                                  'id': k,
                                  'info': [person_image.shape[0], person_image.shape[1], person_image.shape[2]]})
        yield person_image, person_annotation
  
  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    person_file = self._persons_file[id]
    person_image = imread(person_file)
    person_id = self._persons_id[id]
    person_annotation = self._persons_annotation[id]
    person_annotation.update({'category_id': person_id,
                              'id': id,
                              'info': [person_image.shape[0], person_image.shape[1], person_image.shape[2]]})
    return person_image, person_annotation

  def split(self, split_params={}, split_method=''):
    assert (self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold'])
  
    category_ids = None
    if split_method == 'kfold':
      np.random.seed(np.int64(self.seed))
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [i for i in range(len(self.ids))]
  
    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = VGGFace(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids
  
    val_dataset = VGGFace(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids
  
    return train_dataset, val_dataset


# -*- coding: UTF-8 -*-
# @Time : 2018/7/17
# @File : warehouse.py
# @Author: Jian <jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
# import qiniu
# from qiniu import Auth, put_file, etag, urlsafe_base64_encode
from antgo.utils import logger
import os
from six.moves import urllib
import uuid

def qiniu_upload(file_path, bucket='mltalker', out_url_base = 'http://experiment.mltalker.com', max_size=10):
  return 'qiniu:%s/%s' % ('http:127.0.0.1', 'aaa')
  access_key = 'ZSC-X2p4HG5uvEtfmn5fsTZ5nqB3h54oKjHt0tU6'
  secret_key = 'Ya8qYwIDXZn6jSJDMz_ottWWOZqlbV8bDTNfCGO0'
  q = Auth(access_key, secret_key)

  if max_size is not None:
    # check file size
    fsize = os.path.getsize(file_path)
    fsize = fsize / float(1024 * 1024)
    if fsize > max_size:
      logger.error('file size is larger than limit (%dMB)'%max_size)
      return None

  key = file_path.split('/')[-1]
  token = q.upload_token(bucket, key, 3600)
  ret, info = put_file(token, key, file_path)
  if ret['key'] == key and ret['hash'] == etag(file_path):
    logger.info('success to upload')
    return 'qiniu:%s/%s' % (out_url_base, key)

  return None


def qiniu_download(address, file_path):
  access_key = 'ZSC-X2p4HG5uvEtfmn5fsTZ5nqB3h54oKjHt0tU6'
  secret_key = 'Ya8qYwIDXZn6jSJDMz_ottWWOZqlbV8bDTNfCGO0'
  q = Auth(access_key, secret_key)
  if address.startswith('qiniu:'):
    address = address.replace('qiniu:','')
  private_url = q.private_download_url(address, expires=3600)

  if os.path.isdir(file_path):
    file_path = os.path.join(file_path, str(uuid.uuid4()))

  try:
    fpath, _ = urllib.request.urlretrieve(private_url, file_path)
    statinfo = os.stat(fpath)
    size = statinfo.st_size

    if size == 0:
      logger.error('couldnt download data')
      return None

    return file_path
  except:
    logger.error('couldnt download data')
    return None


# import tarfile
# import subprocess
#
# def _recursive_tar(root_path, path, tar, ignore=None):
#   if path.split('/')[-1][0] == '.':
#     return
#
#   if os.path.isdir(path):
#     for sub_path in os.listdir(path):
#       _recursive_tar(root_path, os.path.join(path, sub_path), tar)
#   else:
#     if ignore is not None:
#       if path.split('/')[-1] == ignore:
#         return
#     arcname = os.path.relpath(path, root_path)
#     tar.add(path, arcname=arcname)
#
# random_code_package_name = str(uuid.uuid4())
# code_tar_path = os.path.join('/Users/Jian/Downloads/aaa', '%s_code.tar.gz' % random_code_package_name)
# tar = tarfile.open(code_tar_path, 'w:gz')
# for sub_path in os.listdir('/Users/Jian/Downloads/aaa'):
#   _recursive_tar('/Users/Jian/Downloads/aaa',
#                       os.path.join('/Users/Jian/Downloads/aaa', sub_path),
#                       tar,
#                       ignore='%s_code.tar.gz' % random_code_package_name)
# tar.close()
#
# crypto_code = str(uuid.uuid4())
#
# crypto_shell = 'openssl enc -e -aes256 -in %s -out %s -k %s' % (
# '%s_code.tar.gz' % random_code_package_name,
# '%s_code_ssl.tar.gz' % random_code_package_name,
# crypto_code)
# subprocess.call(crypto_shell, shell=True, cwd='/Users/Jian/Downloads/aaa')
#
# # 解密
# decrypto_shell = 'openssl enc -d -aes256 -in %s -out %s -k %s'%('%s_code_ssl.tar.gz' % random_code_package_name,
#                                                                               '%s_code.tar.gz' % random_code_package_name,
#                                                                               crypto_code)
# subprocess.call(decrypto_shell, shell=True, cwd='/Users/Jian/Downloads/aaa')

# # qiniu_address = qiniu_upload(code_tar_path, bucket='mltalker', max_size=10)
# # print(qiniu_address)
#
# qiniu_download('http://pbzz7zw0y.bkt.clouddn.com/14697433-5bd9-4b3c-a392-ba72172c666e_code.tar.gz', '/Users/Jian/Downloads/zj')


# -*- coding: UTF-8 -*-
# @Time    : 17-6-22
# @File    : trainer.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import re
from abc import ABCMeta, abstractmethod
from antgo.context import *
import collections


DefaultParam = collections.namedtuple('DefaultParam', ['name', 'value', 'help'])


trainer_default_params = [
  DefaultParam('batch_size', 1, 'The number of samples in each batch.'),
  DefaultParam('max_epochs', 100, ''),
  DefaultParam('num_samples', 0, ''),
  DefaultParam('snapshot_prefix', 'alpha', ''),
  DefaultParam('snapshot_infix', 'train', ''),
  DefaultParam('num_clones', 1, 'Number of model clones to deploy'),
  DefaultParam('devices', [], 'clone is deployed at devices'),
  DefaultParam('clone_on_cpu', False, 'Use CPUs to deploy clones'),
  DefaultParam('worker_replicas', 1, 'Number of worker replicas'),
  DefaultParam('num_ps_tasks', 0, 'The number of parameter servers. If the value is 0, then the parameters are handled locally by the worker.'),
  DefaultParam('is_distribute_training', False, 'training model in multi computers'),
  DefaultParam('log_every_n_steps', 50, 'The frequency with which logs are print'),
  DefaultParam('replica_id', 0, 'Task id of the replica running the training.'),
  DefaultParam('weight_decay', 0.00004, 'The weight decay on the model weights.'),
  DefaultParam('optimizer', 'rmsprop', 'The name of the optimizer, one of "adadelta", "adagrad", "adam", "ftrl", "momentum", "sgd" or "rmsprop".'),
  DefaultParam('adadelta_rho', 0.95, 'The decay rate for adadelta.'),
  DefaultParam('adagrad_initial_accumulator_value', 0.1, 'Starting value for the AdaGrad accumulators'),
  DefaultParam('adam_beta1', 0.9, 'The exponential decay rate for the 1st moment estimates.'),
  DefaultParam('adam_beta2', 0.999, 'The exponential decay rate for the 2nd moment estimates'),
  DefaultParam('opt_epsilon', 1e-10, 'Epsilon term for the optimizer.'),
  DefaultParam('ftrl_learning_rate_power', -0.5, 'The learning rate power.'),
  DefaultParam('ftrl_initial_accumulator_value', 0.1, 'Starting value for the FTRL accumulators.'),
  DefaultParam('ftrl_l1', 0.0, 'The FTRL l1 regularization strength'),
  DefaultParam('ftrl_l2', 0.0, 'The FTRL l2 regularization strength.'),
  DefaultParam('momentum', 0.9, 'The momentum for the MomentumOptimizer and RMSPropOptimizer.'),
  DefaultParam('rmsprop_momentum', 0.0, 'Momentum'),
  DefaultParam('rmsprop_decay', 0.9, 'Decay term for RMSProp.'),
  DefaultParam('learning_rate_decay_type', 'polynomial', 'Specifies how the learning rate is decayed. One of "fixed", "exponential", or "polynomial"'),
  DefaultParam('learning_rate', 0.01, 'Initial learning rate'),
  DefaultParam('end_learning_rate', 0.0001, 'The minimal end learning rate used by a polynomial decay learning rate'),
  DefaultParam('learning_rate_decay_factor', 0.94, 'Learning rate decay factor'),
  DefaultParam('num_epochs_per_decay', 20, 'Number of epochs after which learning rate decays.'),
  DefaultParam('sync_replicas', False, 'Whether or not to synchronize the replicas during training.'),
  DefaultParam('replicas_to_aggregate', 1, 'The Number of gradients to collect before updating params.'),
  DefaultParam('moving_average_decay', None, 'The decay to use for the moving average.'),
  DefaultParam('checkpoint_path', None, ''),
  DefaultParam('checkpoint_exclude_scopes', None, ''),
  DefaultParam('checkpoint_fuzzy_exclude_scopes', None, ''),
  DefaultParam('checkpoint_transfer_scopes', None, ''),
  DefaultParam('auxilary_checkpoints',{},''),
  DefaultParam('regularization_loss', True, ''),
  DefaultParam('trainable_filter', None, ''),
  DefaultParam('dataset_queue_threads', 1, ''),
  DefaultParam('trainable_scopes', None, ''),
  DefaultParam('trainable_filter', None, ''),
  DefaultParam('max_no_improvement_num', 5, ''),
  DefaultParam('min_loss_dec', 1e-4, ''),
]


class Trainer(object):
    def __init__(self, is_training=True):
        trainer_context = get_global_context()
        assert(trainer_context is not None)
        # 1.step config trainer context
        for param in trainer_default_params:
          k = param.name
          v = param.value
          if trainer_context is not None:
              setattr(self, k, getattr(trainer_context.params, k, v))
          else:
              setattr(self, k, v)

        self._trainer_context = trainer_context

        # 3.step other
        self.iter = 0
        self.is_training = is_training

        # context
        self.ctx = get_global_context()
        self.ctx.registry_clear_callback(self.wait_until_clear)

        #
        self.early_stop = EarlyStop(self.max_no_improvement_num, self.min_loss_dec, self.ctx)
        
    def deploy(self, model):
        pass

    def run(self, data_generator, binds):
        if self.ctx is not None:
            for k, v, c, f in self.ctx.registried_trainer_callbacks:
                cur_value = getattr(self, k, None)
                if cur_value is not None:
                    if c == 'equal':
                        if cur_value == v:
                            f()
                    elif c == 'less':
                        if cur_value < v:
                            f()
                    elif c == 'greater':
                        if cur_value > v:
                            f()
                    elif c == 'mod':
                        if int(cur_value) % int(v) == 0:
                            f()

    def snapshot(self, dump_dir, epoch):
        pass

    def watch(self, name, fuzzy=True):
        # add watch var list
        pass

    @property
    def iter_at(self):
        return self.iter
    @iter_at.setter
    def iter_at(self, val):
        self.iter = val

    def wait_until_clear(self):
        pass


class ModelDesc(object):
  __metaclass__ = ABCMeta

  def __init__(self, model_name=None, model_data_source=None):
    if model_name is not None:
      self.model_name = model_name
    else:
      self.model_name = self.__class__.__name__
  
    self._ctx = get_global_context()
    self._data_source = model_data_source
    self._model_variables = None

    self._watch_vars = {}
    self._trainer = None

  @property
  def name(self):
    return self.model_name

  @property
  def trainer(self):
    return self._trainer
  @trainer.setter
  def trainer(self, val):
    self._trainer = val

  # @property
  # def ctx(self):
  #   return self._ctx
  # @ctx.setter
  # def ctx(self, val):
  #   self._ctx = val
  #   for k in dir(self._ctx):
  #     if not k.startswith('__'):
  #       setattr(self, k, getattr(self._ctx, k, None))

  @property
  def need_feed(self):
    return self._need_feed
  @need_feed.setter
  def need_feed(self, val):
    self._need_feed = val
  
  @property
  def data_source(self):
    return self._data_source
  
  def model_input(self, is_training, *args, **kwargs):
    return None
  
  @abstractmethod
  def model_fn(self, is_training=True, *args, **kwargs):
    '''
    :return: 
    '''

  @property
  def model_variables(self):
      return self._model_variables
  @model_variables.setter
  def model_variables(self, val):
      self._model_variables = val

  @property
  def automl(self):
    return getattr(self._ctx.params, 'automl', None)

  def watch(self, **kwargs):
    for k,v in kwargs.items():
      self._watch_vars[k] = v

  def get_watch(self, *args):
    a = []
    for k in args:
      if k in self._watch_vars:
        a.append(self._watch_vars[k])
      else:
        a.append(None)
    return a


class EarlyStop(object):
  def __init__(self,
               max_no_improvement_num,
               min_loss_dec,
               ctx=None):
    super(EarlyStop, self).__init__()
    self.training_losses = []
    self.minimum_loss = None
    self.no_improvement_count = 0
    self._max_no_improvement_num = max_no_improvement_num
    self._done = False
    self._min_loss_dec = min_loss_dec
    self._ctx = ctx
    self._is_stop = False
    self.on_train_begin()

  @contextmanager
  def monitor(self, loss_transfer=lambda x: 1.0 - x):
    yield self

    if self._ctx is not None and self._ctx.recorder is not None:
      evaluation_val = self._ctx.recorder.finish()
      is_continue = self.on_epoch_end(loss_transfer(evaluation_val))
      if not is_continue:
        self._is_stop = True

  def on_train_begin(self):
    self.training_losses = []
    self.no_improvement_count = 0
    self._done = False
    self.minimum_loss = float('inf')

  def on_epoch_end(self, loss):
    self.training_losses.append(loss)
    if self._done and loss >= (self.minimum_loss - self._min_loss_dec):
      return False

    if loss >= (self.minimum_loss - self._min_loss_dec):
      self.no_improvement_count += 1
    else:
      self.no_improvement_count = 0
      self.minimum_loss = loss

    if self.no_improvement_count > self._max_no_improvement_num:
      self._done = True

    return True

  @property
  def is_stop(self):
    return self._is_stop


# encoding=utf-8
# @Time    : 17-8-2
# @File    : standard.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.dataflow.dataset.dataset import *
from antgo.dataflow.core import *
from antgo.utils.fs import maybe_here_match_format
from antgo.dataflow.basic import *
import numpy as np
import copy
import time
from antgo.utils.dht import *
import multiprocessing

class Standard(Dataset):
  is_complete = {}

  def __init__(self, train_or_test, dataset_dir=None, ext_params=None):
    dataset_name = dataset_dir.split('/')[-1]
    super(Standard, self).__init__(train_or_test, dataset_dir, ext_params, dataset_name)

    # read sample data
    if self.train_or_test == 'sample':
      self.data_samples, self.ids = self.load_samples()
      return

    # maybe dataset has not prepared now
    if not os.path.exists(os.path.join(dataset_dir,train_or_test)):
      os.makedirs(os.path.join(dataset_dir,train_or_test))

    # data queue (provide data from another independent process)
    self.data_queue = None
    dataset_url = getattr(self, 'dataset_url', None)
    if dataset_url is not None and dataset_url != 'http://xxx.xxx.xxx.com':
      if '%s_%s'%(dataset_name, train_or_test) not in Standard.is_complete:
        logger.info('download %s (%s dataset) from dht asynchronously'%(dataset_name, train_or_test))
        # set complete flag
        Standard.is_complete['%s_%s' % (dataset_name, train_or_test)] = False
        # db reader
        self._record_reader = RecordReader(os.path.join(dataset_dir, train_or_test),
                                           read_only=False)

        # launch independent process
        dataset_url = dataset_url.replace('ipfs://','')
        self.data_queue = multiprocessing.Queue()
        dht_process = multiprocessing.Process(target=dataset_download_dht,
                                              args=(dataset_dir,
                                                    train_or_test,
                                                    dataset_url,
                                                    self.data_queue,
                                                    self._record_reader,
                                                    2))
        dht_process.start()
    else:
      Standard.is_complete['%s_%s' % (dataset_name, train_or_test)] = True
      # db reader
      self._record_reader = RecordReader(os.path.join(dataset_dir, train_or_test),
                                         read_only=True)

    # dataset basic property
    dataset_attrs = {}
    if self.data_queue is not None:
      dataset_attrs = self.data_queue.get()
    else:
      dataset_attrs = self._record_reader.record_attrs()

    for k, v in dataset_attrs.items():
      setattr(self, k, v)

    # dataset index
    self.ids = np.arange(0, int(self.count)).tolist()

    if Standard.is_complete['%s_%s'%(dataset_name, train_or_test)]:
      self.ok_ids = np.arange(0, int(self.size)).tolist()
    else:
      self.ok_ids = []

    # fixed seed
    self.seed = time.time()

  def data_pool(self):
    if self.train_or_test == 'sample':
      sample_idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(sample_idxs)

      for index in sample_idxs:
        yield self.data_samples[index]
      return

    self.epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if self.epoch >= max_epoches:
        break

      if not Standard.is_complete['%s_%s'%(self.name, self.train_or_test)]:
        # for !train dataset, we have to wait until complete
        while True:
          # receive data list
          data_list = self.data_queue.get()
          for data in data_list:
            if type(data) == tuple:
              data_index, data_info, data_label, block_id = data
              # record sample
              self._record_reader.write(Sample(data=data_info, label=data_label), data_index)
              # record block
              self._record_reader.put(block_id, 'true')

              # update ids (sample index)
              self.ok_ids.append(data_index)
            else:
              # update ids (sample index)
              self.ok_ids.append(int(data))

          # check whether dataset is complete
          if self.size == len(self.ok_ids):
            Standard.is_complete['%s_%s' % (self.name, self.train_or_test)] = True
            break

          if self.train_or_test == 'train':
            break

      ids = copy.copy(self.ids)
      if self.rng:
        self.rng.shuffle(ids)

      # filter by ids
      filter_ids = getattr(self, 'filter', None)
      if filter_ids is not None:
        ids = [i for i in ids if i in filter_ids]

      for id in ids:
        if not Standard.is_complete['%s_%s'%(self.name, self.train_or_test)]:
          if id not in self.ok_ids:
            continue

        data, label = self._record_reader.read(id, 'data', 'label')
        # print((id, data,label))
        # filter by condition
        if type(label) == dict:
          if 'category' in label and 'category_id' in label:
            label = self.filter_by_condition(label)
            if label is None:
              continue
          
          label['id'] = id
        yield [data, label]

      # increment epoch
      self.epoch += 1

  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')

    val_dataset = Standard('val', self.dir, self.ext_params)
    return self, val_dataset

  @property
  def size(self):
    if self.train_or_test == 'sample':
      return len(self.ids)

    return int(self.count)
  
  def at(self, id):
    if self.train_or_test == 'sample':
      return self.data_samples[id]

    data, label = self._record_reader.read(id, 'data', 'label')
    return data, label


# -*- coding: UTF-8 -*-
# @Time    : 2020-06-10 12:18
# @File    : base_server.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import tornado.httpserver
import tornado.ioloop
import tornado.options
from tornado.options import define, options
from tornado import web, gen
from tornado import httpclient
import tornado.web
import json


# error code
class RESPONSE_STATUS_CODE:
    SUCCESS = 200                   # [GET]
    RESOURCE_SUCCESS_CREATED = 201  # 新建、修改数据成功 [POST/PUT/PATCH]
    RESOURCE_SUCCESS_DELETE = 204   # 数据删除成功
    RESOURCE_NOT_FOUND = 404        # 未获得指定数据
    RESOURCE_GONE = 410             # 请求的资源已经被永久删除，不可得到 [GET]
    EXECUTE_ACCEPTED = 202          # 任务等待执行
    EXECUTE_UNAUTHORIZED = 401      # 未授权执行
    EXECUTE_FORBIDDEN = 403         # 拥有授权，但访问禁止
    REQUEST_NOT_ACCEPTED = 406      # 用户请求格式错误 [GET]
    REQUEST_INVALID = 400           # 用户请求数据存在问题 [POST/PUT/PATCH]
    INTERNAL_SERVER_ERROR = 500     # 服务器发生错误


class BaseHandler(tornado.web.RequestHandler):
  """Base Handler class with access to common methods and properties."""

  def set_default_headers(self):
    self.set_header("Access-Control-Allow-Origin", "*")
    self.set_header('Access-Control-Allow-Headers', '*')
    self.set_header("Access-Control-Allow-Methods", "POST,GET,PUT,DELETE,PATCH,OPTIONS")
    self.set_header("Access-Control-Expose-Headers", "Content-Disposition")
    # self.set_header("Access-Control-Allow-Credentials", "true")

  @property
  def name(self):
    return self.settings.get('name', '-')

  @property
  def html_template(self):
    return self.settings['html_template']

  @property
  def keywords_template(self):
    return self.settings.get('keywords_template', {})

  @property
  def data_folder(self):
    return self.settings['data_folder']

  @property
  def db(self):
    return self.settings['db']

  @property
  def token(self):
    return self.settings['token']

  @property
  def task_name(self):
    return self.settings['task_name']

  @property
  def task_type(self):
    return self.settings['task_type']

  @property
  def upload_folder(self):
    return self.settings.get('upload', None)

  @property
  def download_folder(self):
    return self.settings.get('download', None)

  @property
  def dump_folder(self):
    return self.settings.get('dump_path', None)

  @property
  def request_queue(self):
    return self.settings.get('request_queue', None)

  @property
  def response_queue(self):
    return self.settings.get('response_queue', None)

  def response(self, status_code=RESPONSE_STATUS_CODE.SUCCESS, message='', content={}, status=None):
    self.set_status(status_code)

    if status is None:
      status = 'OK'
      if status_code not in [200, 201, 204]:
        status = 'ERROR'

    self.write(json.dumps({
      'status': status,
      'message': message,
      'content': content
    }))


# -*- coding: UTF-8 -*-
# Time: 1/6/18
# File: __init__.py.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function



# encoding=utf-8
# @Time    : 17-5-4
# @File    : multic_task.py
# @Author  :
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
from antgo.measures.multi_c import *
from antgo.measures.confusion_matrix import *
import numpy as np


default = {'AntAccuracyMultiC': ('ACCURACY', 'CLASSIFICATION'),
           'AntConfusionMatrixMultiC': ('CONFUSION-MATRIX', 'CLASSIFICATION')}
class AntAccuracyMultiC(AntMeasure):
  def __init__(self, task):
    super(AntAccuracyMultiC, self).__init__(task, 'ACCURACY')
    assert(task.task_type == 'CLASSIFICATION')
    self.is_support_rank = True
    self.larger_is_better = 1

  def eva(self, data, label):
    '''
    :param data: logits (N x class_num)
    :param label: ground truth label (0 ~ ...) (N,)
    :return: accuracy
    '''
    if label is not None:
      data = zip(data, label)

    acutal_label = []
    predicated_label = []
    sample_scores = []
    for predict, gt in data:
      predicated_label.append(predict)

      id = None
      gt_label = gt
      if type(gt) == dict:
        gt_label = int(gt['category_id'])
        id = gt['id']

      if id is not None:
        s = 1 if int(predict) == int(gt_label) else 0
        sample_scores.append({'id': id, 'score': int(s), 'category': gt_label})
      acutal_label.append(gt_label)

    accuracy = multi_accuracy_labels(acutal_label, predicated_label)
    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': float(accuracy), 'type': 'SCALAR'}]},
            'info': sample_scores}

    # curve_data = [[[0,0],[1,1],[2,2]]]
    # bar_data = [10,5,2]
    # return {'statistic': {'name': self.name,
    #                       'value': [{'name': self.name, 'value': float(accuracy), 'type': 'SCALAR'},
    #                                 {'name': 'test-curve', 'value': curve_data,
    #                                  'type': 'CURVE', 'x': 'x', 'y': 'y'},
    #                                 {'name': 'test-bar', 'value': bar_data,
    #                                  'type': 'SCALAR', 'x': ['a', 'b', 'c'], 'y': 'y'}]}}


class AntConfusionMatrixMultiC(AntMeasure):
  def __init__(self, task):
    super(AntConfusionMatrixMultiC, self).__init__(task, 'CONFUSION-MATRIX')
    assert(task.task_type == 'CLASSIFICATION')

  def eva(self, data, label):
    '''
    :param data: logits (N x class_num)
    :param label: ground truth label (0 ~ ...) (N,)
    :return: matrix
    '''
    if label is not None:
      data = zip(data, label)

    acutal_label = []
    predicated_label = []
    for predict, gt in data:
      predicated_label.append(int(predict))

      gt_label = -1
      if type(gt) == dict:
        gt_label = int(gt['category_id'])
      else:
        gt_label = int(gt)

      acutal_label.append(gt_label)

    class_num = len(self.task.class_label)
    cm = compute_confusion_matrix(acutal_label, predicated_label, class_num)

    # ######
    # curve_data = [[[0,0],[1,1],[2,2]]]
    # bar_data = [10,5,2]
    # image = np.random.random((50,50,3))
    # image = image*255
    # image = image.astype(np.uint8)
    # {'name': 'test-curve', 'value': curve_data,
    #  'type': 'CURVE', 'x': 'x', 'y': 'y'},
    # {'name': 'test-bar', 'value': bar_data,
    #  'type': 'SCALAR', 'x': ['a', 'b', 'c'], 'y': 'y'},
    # {'name': 'test-image', 'value': image,
    #  'type': 'IMAGE'}]
    # ######

    return {'statistic': {'name': self.name,
                          'value': [{'name': self.name, 'value': cm.tolist(),
                                     'type': 'MATRIX', 'x': self.task.class_label, 'y': self.task.class_label}
                                    ]}}

    # return {'statistic': {'name': self.name,
    #                       'value': [{'name': self.name, 'value': cm.tolist(),
    #                                  'type': 'MATRIX', 'x': self.task.class_label, 'y': self.task.class_label},
    #                                 {'name': 'test-curve', 'value': curve_data,
    #                                  'type': 'CURVE', 'x': 'x', 'y': 'y'},
    #                                 {'name': 'test-bar', 'value': bar_data,
    #                                 'type': 'SCALAR', 'x': ['a', 'b', 'c'], 'y': 'y'}]}}


# -*- coding: UTF-8 -*-
# @Time    : 2019-04-17 09:38
# @File    : track_task.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.measures.base import *
from antgo.utils._bbox import bbox_overlaps
from antgo.task.task import *

default={'AntTrackStability': ('TrackStability', 'TRACK')}


class AntTrackStability(AntMeasure):
  def __init__(self, task):
    super(AntTrackStability, self).__init__(task, "TrackStability")
    assert(task.task_type == 'TRACK')
    self.iou_thres = 0.5
    self.support_rank_index = 0
    self.is_support_rank = True
    self.is_inverse = True

  def _fragment_error_fast(self, trajectory):
    t_k = len(trajectory)
    if t_k <= 1:
      return [0.0] * 11

    f_k = np.zeros((len(trajectory)))
    scores_k = np.zeros((len(trajectory)))

    index = 0
    for det_bbox, gt_bbox, det_score, is_ok in trajectory:
      scores_k[index] = det_score
      if index == 0:
        f_k[index] = 1.0 if is_ok else 0
      else:
        f_k[index] = f_k[index-1]+1 if is_ok else f_k[index-1]
      index += 1

    sampling_p = np.array(range(11)) * 0.1
    error_k = f_k / (t_k - 1.0)
    error_k = np.interp(sampling_p, scores_k.tolist()[::-1], error_k.tolist()[::-1])
    return error_k

  def _center_position_error_fast(self, trajectory):
    if len(trajectory) <= 1:
      return [0.0] * 11

    e_x = np.zeros((len(trajectory)))
    e_y = np.zeros((len(trajectory)))
    error_k = np.zeros((len(trajectory) - 1))
    scores_k = np.zeros((len(trajectory) - 1))
    index = 0
    for det_bbox, gt_bbox, det_socre, is_ok in trajectory:
      if not is_ok:
        if index > 1:
          error_k[index - 1] = error_k[index - 2]
          scores_k[index - 1] = det_socre
          index += 1
        continue

      e_x[index] = (det_bbox[0] - gt_bbox[0])/(gt_bbox[2]-gt_bbox[0])
      e_y[index] = (det_bbox[1] - gt_bbox[1])/(gt_bbox[3]-gt_bbox[1])
      if index > 0:
        error_k[index-1] = np.std(e_x[0:index+1])+np.std(e_y[0:index+1])
        scores_k[index-1] = det_socre
      index += 1

    sampling_p = np.array(range(11)) * 0.1
    error_k = np.interp(sampling_p, scores_k.tolist()[::-1], error_k.tolist()[::-1])
    return error_k

  def _scale_ratio_error_fast(self, trajectory):
    if len(trajectory) <= 1:
      return [0.0] * 11

    e_s = np.zeros((len(trajectory)))
    e_r = np.zeros((len(trajectory)))
    error_k = np.zeros((len(trajectory) - 1))
    scores_k = np.zeros((len(trajectory) - 1))
    index = 0
    for det_bbox, gt_bbox, det_socre, is_ok in trajectory:
      if not is_ok:
        if index > 1:
          error_k[index - 1] = error_k[index - 2]
          scores_k[index - 1] = det_socre
          index += 1
        continue

      e_s[index] = np.sqrt(((det_bbox[2] - det_bbox[0]) * (det_bbox[3] - det_bbox[1])) / (
              (gt_bbox[2]-gt_bbox[0]) * (gt_bbox[3] - gt_bbox[1])))

      e_r[index] = ((det_bbox[2] - det_bbox[0]) / (det_bbox[3] - det_bbox[1])) / (
                (gt_bbox[2] - gt_bbox[0]) / (gt_bbox[3] - gt_bbox[1]))

      if index > 0:
        error_k[index - 1] = np.std(e_s[0:index+1]) + np.std(e_r[0:index+1])
      scores_k[index - 1] = det_socre
      index += 1

    sampling_p = np.array(range(11)) * 0.1
    error_k = np.interp(sampling_p, scores_k.tolist()[::-1], error_k.tolist()[::-1])
    return error_k

  def eva(self, data, label):
    # video list
    if label is not None:
      data = zip(data, label)

    all_trajectory_list = []
    for video_predict, video_gt in data:
      trajectory_list = []
      for det_predict, det_gt in zip(video_predict, video_gt):
        # predict bbox in frame
        det_bbox = np.array(det_predict['det-bbox'])
        det_socre = np.array(det_predict['det-score'])
        det_label = np.array(det_predict['det-label']).astype(dtype=np.int32)

        # gt bbox in frame
        gt_bbox = det_gt['bbox']
        gt_category = det_gt['category_id']
        gt_bbox_num = len(gt_bbox)
        if len(trajectory_list) == 0:
          trajectory_list = [[] for _ in range(gt_bbox_num)]

        overlaps = bbox_overlaps(
          np.ascontiguousarray(det_bbox, dtype=np.float),
          np.ascontiguousarray(gt_bbox, dtype=np.float))

        gtm = np.ones((gt_bbox_num)) * (-1)
        for dind, d in enumerate(det_bbox.tolist()):
          best_match = -1
          iou_thres = self.iou_thres
          for gind, g in enumerate(gt_bbox):
            if int(gt_category[gind]) != int(det_label[dind]):
              continue

            if gtm[gind] >= 0:
              continue

            if overlaps[dind, gind] < iou_thres:
              continue

            iou_thres = overlaps[dind, gind]
            best_match = gind

          if best_match > -1:
            gtm[best_match] = dind
            trajectory_list[int(gt_category[best_match])].append((det_bbox[dind],
                                                                  gt_bbox[best_match],
                                                                  det_socre[dind],
                                                                  True))
        for gind, g in enumerate(gt_bbox):
            if gtm[gind] == -1:
              trajectory_list[int(gt_category[gind])].append(([],
                                                              gt_bbox[gind],
                                                              0.0,
                                                              False))

      all_trajectory_list.extend(trajectory_list)

    trajectory_error_list = []
    trajectory_f_error_list = []
    trajectory_cp_error_list = []
    trajectory_sr_error_list = []
    for trajectory in all_trajectory_list:
      ordered_trajectory = sorted(trajectory, key=lambda x: x[2], reverse=True)

      error_f = self._fragment_error_fast(ordered_trajectory)
      error_cpe = self._center_position_error_fast(ordered_trajectory)
      error_sre = self._scale_ratio_error_fast(ordered_trajectory)
      trajectory_f_error_list.append(error_f.tolist())
      trajectory_cp_error_list.append(error_cpe.tolist())
      trajectory_sr_error_list.append(error_sre.tolist())
      trajectory_error_list.append((error_f+error_cpe+error_sre).tolist())

    trajectory_error_mean = np.mean(trajectory_error_list, 0)
    trajectory_f_error_mean = np.mean(trajectory_f_error_list, 0)
    trajectory_cp_error_mean = np.mean(trajectory_cp_error_list, 0)
    trajectory_sr_error_mean = np.mean(trajectory_sr_error_list, 0)
    stability_e = np.sum(trajectory_error_mean)

    return {'statistic': {'name': self.name,
                          'value': [{'name': 'TRAJECTORY_STABILITY',
                                     'value': stability_e,
                                     'type': 'SCALAR',
                                     'x': 'class',
                                     'y': 'STABILITY'},
                                    {'name': 'TRAJECTORY_FRAGMENT_ERROR',
                                     'value': trajectory_f_error_mean.tolist(),
                                     'type': 'SCALAR',
                                     'x': 'Detect Threshold',
                                     'y': 'Fragment Error'},
                                    {'name': 'TRAJECTORY_CENTER_POSITION_ERROR',
                                     'value': trajectory_cp_error_mean.tolist(),
                                     'type': 'SCALAR',
                                     'x': 'Detect Threshold',
                                     'y': 'Center Position Error'},
                                    {'name': 'TRAJECTORY_SCALE_RATIO_ERROR',
                                     'value': trajectory_sr_error_mean.tolist(),
                                     'type': 'SCALAR',
                                     'x': 'Detect Threshold',
                                     'y': 'Scale Ratio Error'},
                                    {'name': 'TRAJECTORY_ERROR',
                                      'value': trajectory_error_mean.tolist(),
                                     'type': 'SCALAR',
                                     'x': 'Detect Threshold',
                                     'y': 'Total Error'}]}}


if __name__ == '__main__':
  data = []
  video_clips_num = 10
  for _ in range(video_clips_num):
    det_result = []
    gt_result = []
    for frame_index in range(30):
      # target 1
      x1 = np.random.random()
      y1 = np.random.random()

      x2 = x1 + np.random.random()
      y2 = y1 + np.random.random()

      gt_x1 = x1 + 0.01
      gt_y1 = y1 + 0.1
      gt_x2 = gt_x1 + 0.1
      gt_y2 = gt_y1 + 0.1

      # target 2
      second_x1 = np.random.random()
      second_y1 = np.random.random()

      second_x2 = second_x1 + np.random.random()
      second_y2 = second_y1 + np.random.random()

      gt_second_x1 = second_x1 + 0.01
      gt_second_y1 = second_y1 + 0.1
      gt_second_x2 = gt_second_x1 + 0.1
      gt_second_y2 = gt_second_y1 + 0.2

      det_result.append({'det-bbox': [[x1, y1, x2, y2], [second_x1, second_y1, second_x2, second_y2]],
                         'det-score': [np.random.random(), np.random.random()],
                         'det-label': [0, 1]})

      gt_result.append({'bbox': [[gt_x1, gt_y1, gt_x2, gt_y2],
                                 [gt_second_x1, gt_second_y1, gt_second_x2, gt_second_y2]],
                        'category_id': [0, 1]})

    data.append([det_result, gt_result])

  ts = AntTrackStability(None)
  result = ts.eva(data, None)